import ast
import json
import logging
import os
import sys
import traceback
from datetime import datetime
from typing import Any, Dict, List, Optional, cast

import click
import pyarrow as pa
import pyarrow.flight as fl
from google.protobuf.json_format import Parse

from feast import FeatureStore, FeatureView, utils
from feast.arrow_error_handler import arrow_server_error_handling_decorator
from feast.data_source import DataSource
from feast.feature_logging import FeatureServiceLoggingSource
from feast.feature_view import DUMMY_ENTITY_NAME
from feast.infra.offline_stores.offline_utils import get_offline_store_from_config
from feast.permissions.action import AuthzedAction
from feast.permissions.security_manager import assert_permissions
from feast.permissions.server.arrow import (
    AuthorizationMiddlewareFactory,
    inject_user_details_decorator,
)
from feast.permissions.server.utils import (
    AuthManagerType,
    ServerType,
    init_auth_manager,
    init_security_manager,
    str_to_auth_manager_type,
)
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.saved_dataset import SavedDatasetStorage

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class OfflineServer(fl.FlightServerBase):
    def __init__(
        self,
        store: FeatureStore,
        location: str,
        host: str = "localhost",
        tls_certificates: List = [],
        **kwargs,
    ):
        super(OfflineServer, self).__init__(
            location=location,
            middleware=self.arrow_flight_auth_middleware(
                str_to_auth_manager_type(store.config.auth_config.type)
            ),
            tls_certificates=tls_certificates,
            verify_client=False,  # this is needed for when we don't need mTLS
            **kwargs,
        )
        self._location = location
        # A dictionary of configured flights, e.g. API calls received and not yet served
        self.flights: Dict[str, Any] = {}
        self.store = store
        self.offline_store = get_offline_store_from_config(store.config.offline_store)
        self.host = host
        self.tls_certificates = tls_certificates

    def arrow_flight_auth_middleware(
        self,
        auth_type: AuthManagerType,
    ) -> dict[str, fl.ServerMiddlewareFactory]:
        """
        A dictionary with the configured middlewares to support extracting the user details when the authorization manager is defined.
        The authorization middleware key is `auth`.

        Returns:
            dict[str, fl.ServerMiddlewareFactory]: Optional dictionary of middlewares. If the authorization type is set to `NONE`, it returns an empty dict.
        """

        if auth_type == AuthManagerType.NONE:
            return {}

        return {
            "auth": AuthorizationMiddlewareFactory(),
        }

    @classmethod
    def descriptor_to_key(self, descriptor: fl.FlightDescriptor):
        return (
            descriptor.descriptor_type.value,
            descriptor.command,
            tuple(descriptor.path or tuple()),
        )

    def _make_flight_info(self, key: Any, descriptor: fl.FlightDescriptor):
        if len(self.tls_certificates) != 0:
            location = fl.Location.for_grpc_tls(self.host, self.port)
        else:
            location = fl.Location.for_grpc_tcp(self.host, self.port)
        endpoints = [
            fl.FlightEndpoint(repr(key), [location]),
        ]
        schema = pa.schema([])

        return fl.FlightInfo(schema, descriptor, endpoints, -1, -1)

    @inject_user_details_decorator
    @arrow_server_error_handling_decorator
    def list_flights(self, context: fl.ServerCallContext, criteria: bytes):
        for key, table in self.flights.items():
            if key[1] is not None:
                descriptor = fl.FlightDescriptor.for_command(key[1])
            else:
                descriptor = fl.FlightDescriptor.for_path(*key[2])

            yield self._make_flight_info(key, descriptor)

    @inject_user_details_decorator
    @arrow_server_error_handling_decorator
    def get_flight_info(
        self, context: fl.ServerCallContext, descriptor: fl.FlightDescriptor
    ):
        key = OfflineServer.descriptor_to_key(descriptor)
        if key in self.flights:
            return self._make_flight_info(key, descriptor)
        raise KeyError("Flight not found.")

    # Expects to receive request parameters and stores them in the flights dictionary
    # Indexed by the unique command
    @inject_user_details_decorator
    @arrow_server_error_handling_decorator
    def do_put(
        self,
        context: fl.ServerCallContext,
        descriptor: fl.FlightDescriptor,
        reader: fl.MetadataRecordBatchReader,
        writer: fl.FlightMetadataWriter,
    ):
        key = OfflineServer.descriptor_to_key(descriptor)
        command = json.loads(key[1])
        if "api" in command:
            data = reader.read_all()
            logger.debug(f"do_put: command is{command}, data is {data}")
            self.flights[key] = data

            self._call_api(command["api"], command, key)
        else:
            logger.warning(f"No 'api' field in command: {command}")

    def _call_api(self, api: str, command: dict, key: str):
        assert api is not None, "api can not be empty"

        remove_data = False
        try:
            if api == OfflineServer.offline_write_batch.__name__:
                self.offline_write_batch(command, key)
                remove_data = True
            elif api == OfflineServer.write_logged_features.__name__:
                self.write_logged_features(command, key)
                remove_data = True
            elif api == OfflineServer.persist.__name__:
                self.persist(command, key)
                remove_data = True
            elif api == OfflineServer.validate_data_source.__name__:
                self.validate_data_source(command)
                remove_data = True
        except Exception as e:
            remove_data = True
            logger.exception(e)
            traceback.print_exc()
            raise e
        finally:
            if remove_data:
                # Get service is consumed, so we clear the corresponding flight and data
                del self.flights[key]

    def get_feature_view_by_name(
        self, fv_name: str, name_alias: str, project: str
    ) -> FeatureView:
        """
        Retrieves a feature view by name, including all subclasses of FeatureView.

        Args:
            fv_name: Name of feature view
            name_alias: Alias to be applied to the projection of the registered view
            project: Feast project that this feature view belongs to

        Returns:
            Returns either the specified feature view, or raises an exception if
            none is found
        """
        try:
            fv = self.store.registry.get_feature_view(name=fv_name, project=project)
            if name_alias is not None:
                for fs in self.store.registry.list_feature_services(project=project):
                    for p in fs.feature_view_projections:
                        if p.name_alias == name_alias:
                            logger.debug(
                                f"Found matching FeatureService {fs.name} with projection {p}"
                            )
                            fv = fv.with_projection(p)
            return fv
        except Exception:
            try:
                return self.store.registry.get_stream_feature_view(
                    name=fv_name, project=project
                )
            except Exception as e:
                logger.error(
                    f"Cannot find any FeatureView by name {fv_name} in project {project}"
                )
                raise e

    def list_feature_views_by_name(
        self, feature_view_names: List[str], name_aliases: List[str], project: str
    ) -> List[FeatureView]:
        return [
            remove_dummies(
                self.get_feature_view_by_name(
                    fv_name=fv_name, name_alias=name_aliases[index], project=project
                )
            )
            for index, fv_name in enumerate(feature_view_names)
        ]

    def _validate_do_get_parameters(self, command: dict):
        assert "api" in command, "api parameter is mandatory"

    # Extracts the API parameters from the flights dictionary, delegates the execution to the FeatureStore instance
    # and returns the stream of data
    @inject_user_details_decorator
    @arrow_server_error_handling_decorator
    def do_get(self, context: fl.ServerCallContext, ticket: fl.Ticket):
        key = ast.literal_eval(ticket.ticket.decode())
        if key not in self.flights:
            logger.error(f"Unknown key {key}")
            return None

        command = json.loads(key[1])

        self._validate_do_get_parameters(command)

        api = command["api"]
        logger.debug(f"get command is {command}")
        logger.debug(f"requested api is {api}")
        try:
            if api == OfflineServer.get_historical_features.__name__:
                table = self.get_historical_features(command, key).to_arrow()
            elif api == OfflineServer.pull_all_from_table_or_query.__name__:
                table = self.pull_all_from_table_or_query(command).to_arrow()
            elif api == OfflineServer.pull_latest_from_table_or_query.__name__:
                table = self.pull_latest_from_table_or_query(command).to_arrow()
            elif (
                api
                == OfflineServer.get_table_column_names_and_types_from_data_source.__name__
            ):
                table = self.get_table_column_names_and_types_from_data_source(command)
            else:
                raise NotImplementedError
        except Exception as e:
            logger.exception(e)
            traceback.print_exc()
            raise e

        # Get service is consumed, so we clear the corresponding flight and data
        del self.flights[key]
        return fl.RecordBatchStream(table)

    def _validate_offline_write_batch_parameters(self, command: dict):
        assert "feature_view_names" in command, (
            "feature_view_names is a mandatory parameter"
        )
        assert "name_aliases" in command, "name_aliases is a mandatory parameter"

        feature_view_names = command["feature_view_names"]
        assert len(feature_view_names) == 1, (
            "feature_view_names list should only have one item"
        )

        name_aliases = command["name_aliases"]
        assert len(name_aliases) == 1, "name_aliases list should only have one item"

    def offline_write_batch(self, command: dict, key: str):
        self._validate_offline_write_batch_parameters(command)

        feature_view_names = command["feature_view_names"]
        name_aliases = command["name_aliases"]

        project = self.store.config.project
        feature_views = self.list_feature_views_by_name(
            feature_view_names=feature_view_names,
            name_aliases=name_aliases,
            project=project,
        )

        assert len(feature_views) == 1, "incorrect feature view"
        table = self.flights[key]
        self.offline_store.offline_write_batch(
            self.store.config,
            cast(
                FeatureView,
                assert_permissions(
                    feature_views[0], actions=[AuthzedAction.WRITE_OFFLINE]
                ),
            ),
            table,
            command["progress"],
        )

    def _validate_write_logged_features_parameters(self, command: dict):
        assert "feature_service_name" in command

    def write_logged_features(self, command: dict, key: str):
        self._validate_write_logged_features_parameters(command)
        table = self.flights[key]
        feature_service = self.store.get_feature_service(
            command["feature_service_name"]
        )

        assert feature_service.logging_config is not None, (
            "feature service must have logging_config set"
        )

        assert_permissions(
            resource=feature_service,
            actions=[AuthzedAction.WRITE_OFFLINE],
        )
        self.offline_store.write_logged_features(
            config=self.store.config,
            data=table,
            source=FeatureServiceLoggingSource(
                feature_service, self.store.config.project
            ),
            logging_config=feature_service.logging_config,
            registry=self.store.registry,
        )

    def _validate_pull_all_from_table_or_query_parameters(self, command: dict):
        assert "data_source_name" in command, (
            "data_source_name is a mandatory parameter"
        )
        assert "join_key_columns" in command, (
            "join_key_columns is a mandatory parameter"
        )
        assert "feature_name_columns" in command, (
            "feature_name_columns is a mandatory parameter"
        )
        assert "timestamp_field" in command, "timestamp_field is a mandatory parameter"
        assert "start_date" in command, "start_date is a mandatory parameter"
        assert "end_date" in command, "end_date is a mandatory parameter"

    def pull_all_from_table_or_query(self, command: dict):
        self._validate_pull_all_from_table_or_query_parameters(command)
        data_source = self.store.get_data_source(command["data_source_name"])
        assert_permissions(data_source, actions=[AuthzedAction.READ_OFFLINE])

        return self.offline_store.pull_all_from_table_or_query(
            config=self.store.config,
            data_source=data_source,
            join_key_columns=command["join_key_columns"],
            feature_name_columns=command["feature_name_columns"],
            created_timestamp_column=command["created_timestamp_column"],
            timestamp_field=command["timestamp_field"],
            start_date=utils.make_tzaware(
                datetime.fromisoformat(command["start_date"])
            ),
            end_date=utils.make_tzaware(datetime.fromisoformat(command["end_date"])),
        )

    def _validate_pull_latest_from_table_or_query_parameters(self, command: dict):
        assert "data_source_name" in command, (
            "data_source_name is a mandatory parameter"
        )
        assert "join_key_columns" in command, (
            "join_key_columns is a mandatory parameter"
        )
        assert "feature_name_columns" in command, (
            "feature_name_columns is a mandatory parameter"
        )
        assert "timestamp_field" in command, "timestamp_field is a mandatory parameter"
        assert "start_date" in command, "start_date is a mandatory parameter"
        assert "end_date" in command, "end_date is a mandatory parameter"

    def pull_latest_from_table_or_query(self, command: dict):
        self._validate_pull_latest_from_table_or_query_parameters(command)
        data_source = self.store.get_data_source(command["data_source_name"])
        assert_permissions(resource=data_source, actions=[AuthzedAction.READ_OFFLINE])
        return self.offline_store.pull_latest_from_table_or_query(
            self.store.config,
            data_source,
            command["join_key_columns"],
            command["feature_name_columns"],
            command["timestamp_field"],
            command["created_timestamp_column"],
            utils.make_tzaware(datetime.fromisoformat(command["start_date"])),
            utils.make_tzaware(datetime.fromisoformat(command["end_date"])),
        )

    @arrow_server_error_handling_decorator
    def list_actions(self, context):
        return [
            (
                OfflineServer.offline_write_batch.__name__,
                "Writes the specified arrow table to the data source underlying the specified feature view.",
            ),
            (
                OfflineServer.write_logged_features.__name__,
                "Writes logged features to a specified destination in the offline store.",
            ),
            (
                OfflineServer.persist.__name__,
                "Synchronously executes the underlying query and persists the result in the same offline store at the "
                "specified destination.",
            ),
        ]

    def _validate_get_historical_features_parameters(
        self, command: dict, key: Optional[str] = None
    ):
        if key:
            assert key in self.flights, f"missing key={key}"
        assert "feature_view_names" in command, "feature_view_names is mandatory"
        assert "name_aliases" in command, "name_aliases is mandatory"
        assert "feature_refs" in command, "feature_refs is mandatory"
        assert "project" in command, "project is mandatory"
        assert "full_feature_names" in command, "full_feature_names is mandatory"

    def get_historical_features(self, command: dict, key: Optional[str] = None):
        self._validate_get_historical_features_parameters(command, key)
        entity_df = None
        if key:
            # Extract parameters from the internal flights dictionary
            entity_df_value = self.flights[key]
            entity_df = pa.Table.to_pandas(entity_df_value)

        feature_view_names = command["feature_view_names"]
        name_aliases = command["name_aliases"]
        feature_refs = command["feature_refs"]
        project = command["project"]
        full_feature_names = command["full_feature_names"]

        feature_views = self.list_feature_views_by_name(
            feature_view_names=feature_view_names,
            name_aliases=name_aliases,
            project=project,
        )

        for feature_view in feature_views:
            assert_permissions(
                resource=feature_view, actions=[AuthzedAction.READ_OFFLINE]
            )

        retJob = self.offline_store.get_historical_features(
            config=self.store.config,
            feature_views=feature_views,
            feature_refs=feature_refs,
            entity_df=entity_df,
            registry=self.store.registry,
            project=project,
            full_feature_names=full_feature_names,
        )

        return retJob

    def _validate_persist_parameters(self, command: dict):
        assert "retrieve_func" in command, "retrieve_func is mandatory"
        assert "data_source_name" in command, "data_source_name is mandatory"
        assert "allow_overwrite" in command, "allow_overwrite is mandatory"

    def persist(self, command: dict, key: str):
        self._validate_persist_parameters(command)

        try:
            retrieve_func = command["retrieve_func"]
            if retrieve_func == OfflineServer.get_historical_features.__name__:
                ret_job = self.get_historical_features(command, key)
            elif (
                retrieve_func == OfflineServer.pull_latest_from_table_or_query.__name__
            ):
                ret_job = self.pull_latest_from_table_or_query(command)
            elif retrieve_func == OfflineServer.pull_all_from_table_or_query.__name__:
                ret_job = self.pull_all_from_table_or_query(command)
            else:
                raise NotImplementedError

            data_source = self.store.get_data_source(command["data_source_name"])
            assert_permissions(
                resource=data_source,
                actions=[AuthzedAction.WRITE_OFFLINE],
            )
            storage = SavedDatasetStorage.from_data_source(data_source)
            ret_job.persist(storage, command["allow_overwrite"], command["timeout"])
        except Exception as e:
            logger.exception(e)
            traceback.print_exc()
            raise e

    @staticmethod
    def _extract_data_source_from_command(command) -> DataSource:
        data_source_proto_str = command["data_source_proto"]
        logger.debug(f"Extracted data_source_proto {data_source_proto_str}")
        data_source_proto = DataSourceProto()
        Parse(data_source_proto_str, data_source_proto)
        data_source = DataSource.from_proto(data_source_proto)
        logger.debug(f"Converted to DataSource {data_source}")
        return data_source

    def validate_data_source(self, command: dict):
        data_source = OfflineServer._extract_data_source_from_command(command)
        logger.debug(f"Validating data source {data_source.name}")
        assert_permissions(data_source, actions=[AuthzedAction.READ_OFFLINE])

        self.offline_store.validate_data_source(
            config=self.store.config,
            data_source=data_source,
        )

    def get_table_column_names_and_types_from_data_source(self, command: dict):
        data_source = OfflineServer._extract_data_source_from_command(command)
        logger.debug(f"Fetching table columns metadata from {data_source.name}")
        assert_permissions(data_source, actions=[AuthzedAction.READ_OFFLINE])

        column_names_and_types = data_source.get_table_column_names_and_types(
            self.store.config
        )

        column_names, types = zip(*column_names_and_types)
        logger.debug(
            f"DataSource {data_source.name} has columns {column_names} with types {types}"
        )
        return pa.table({"name": column_names, "type": types})

    def serve(self):
        message = "offline server starting with pid: "
        logger.info(
            message + "[%d]",
            os.getpid(),
            extra={"color_message": message + "[" + click.style("%d", fg="cyan") + "]"},
        )
        super().serve()

    def shutdown(self):
        message = "Sending a shutdown signal to the offline server running with pid:: "
        logger.info(
            message + "[%d]",
            os.getpid(),
            extra={"color_message": message + "[" + click.style("%d", fg="cyan") + "]"},
        )
        super().shutdown()


def remove_dummies(fv: FeatureView) -> FeatureView:
    """
    Removes dummmy IDs from FeatureView instances created with FeatureView.from_proto
    """
    if DUMMY_ENTITY_NAME in fv.entities:
        fv.entities = []
        fv.entity_columns = []
    return fv


def _init_auth_manager(store: FeatureStore):
    auth_type = str_to_auth_manager_type(store.config.auth_config.type)
    init_security_manager(auth_type=auth_type, fs=store)
    init_auth_manager(
        auth_type=auth_type,
        server_type=ServerType.ARROW,
        auth_config=store.config.auth_config,
    )


def start_server(
    store: FeatureStore,
    host: str,
    port: int,
    tls_key_path: str = "",
    tls_cert_path: str = "",
):
    _init_auth_manager(store)

    tls_certificates = []
    scheme = "grpc+tcp"
    if tls_key_path and tls_cert_path:
        logger.info(
            "Found SSL certificates in the args so going to start offline server in TLS(SSL) mode."
        )
        scheme = "grpc+tls"
        with open(tls_cert_path, "rb") as cert_file:
            tls_cert_chain = cert_file.read()
        with open(tls_key_path, "rb") as key_file:
            tls_private_key = key_file.read()
        tls_certificates.append((tls_cert_chain, tls_private_key))

    location = "{}://{}:{}".format(scheme, host, port)
    server = OfflineServer(
        store,
        location=location,
        host=host,
        tls_certificates=tls_certificates,
    )
    try:
        logger.info(f"Offline store server serving at: {location}")
        server.serve()
    except KeyboardInterrupt:
        logger.info("KeyboardInterrupt received, stopping the offline server.")
    finally:
        server.shutdown()
        logger.info("offline server stopped.")
        sys.exit(0)



from typing import Any, Optional

from feast.infra.compute_engines.dag.model import DAGFormat


class DAGValue:
    def __init__(self, data: Any, format: DAGFormat, metadata: Optional[dict] = None):
        self.data = data
        self.format = format
        self.metadata = metadata or {}

    def assert_format(self, expected: DAGFormat):
        if self.format != expected:
            raise ValueError(f"Expected format {expected}, but got {self.format}")



import functools
import warnings
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import dill

from feast import flags_helper
from feast.aggregation import Aggregation
from feast.data_source import DataSource
from feast.entity import Entity
from feast.feature_view import FeatureView
from feast.field import Field
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.transformation.base import Transformation
from feast.transformation.mode import TransformationMode

warnings.simplefilter("once", RuntimeWarning)

SUPPORTED_BATCH_SOURCES = {
    "BigQuerySource",
    "FileSource",
    "RedshiftSource",
    "SnowflakeSource",
    "SparkSource",
    "TrinoSource",
    "AthenaSource",
}


class BatchFeatureView(FeatureView):
    """
    A batch feature view defines a logical group of features that has only a batch data source.

    Attributes:
        name: The unique name of the batch feature view.
        mode: The transformation mode to use for the batch feature view. This can be one of TransformationMode
        entities: List of entities or entity join keys.
        ttl: The amount of time this group of features lives. A ttl of 0 indicates that
            this group of features lives forever. Note that large ttl's or a ttl of 0
            can result in extremely computationally intensive queries.
        schema: The schema of the feature view, including feature, timestamp, and entity
            columns. If not specified, can be inferred from the underlying data source.
        source: The batch source of data where this group of features is stored.
        online: A boolean indicating whether online retrieval and write to online store is enabled for this feature view.
        offline: A boolean indicating whether offline retrieval and write to offline store is enabled for this feature view.
        description: A human-readable description.
        tags: A dictionary of key-value pairs to store arbitrary metadata.
        owner: The owner of the batch feature view, typically the email of the primary maintainer.
        udf: A user-defined function that applies transformations to the data in the batch feature view.
        udf_string: A string representation of the user-defined function.
        feature_transformation: A transformation object that defines how features are transformed.
                Note, feature_transformation has precedence over udf and udf_string.
        batch_engine: A dictionary containing configuration for the batch engine used to process the feature view.
                Note, it will override the repo-level default batch engine config defined in the yaml file.
        aggregations: A list of aggregations to be applied to the features in the batch feature view.
    """

    name: str
    mode: Union[TransformationMode, str]
    entities: List[str]
    ttl: Optional[timedelta]
    source: DataSource
    sink_source: Optional[DataSource] = None
    schema: List[Field]
    entity_columns: List[Field]
    features: List[Field]
    online: bool
    offline: bool
    description: str
    tags: Dict[str, str]
    owner: str
    timestamp_field: str
    materialization_intervals: List[Tuple[datetime, datetime]]
    udf: Optional[Callable[[Any], Any]]
    udf_string: Optional[str]
    feature_transformation: Optional[Transformation]
    batch_engine: Optional[Dict[str, Any]]
    aggregations: Optional[List[Aggregation]]

    def __init__(
        self,
        *,
        name: str,
        mode: Union[TransformationMode, str] = TransformationMode.PYTHON,
        source: Union[DataSource, "BatchFeatureView", List["BatchFeatureView"]],
        sink_source: Optional[DataSource] = None,
        entities: Optional[List[Entity]] = None,
        ttl: Optional[timedelta] = None,
        tags: Optional[Dict[str, str]] = None,
        online: bool = False,
        offline: bool = False,
        description: str = "",
        owner: str = "",
        schema: Optional[List[Field]] = None,
        udf: Optional[Callable[[Any], Any]] = None,
        udf_string: Optional[str] = "",
        feature_transformation: Optional[Transformation] = None,
        batch_engine: Optional[Dict[str, Any]] = None,
        aggregations: Optional[List[Aggregation]] = None,
    ):
        if not flags_helper.is_test():
            warnings.warn(
                "Batch feature views are experimental features in alpha development. "
                "Some functionality may still be unstable so functionality can change in the future.",
                RuntimeWarning,
            )

        if isinstance(source, DataSource) and (
            type(source).__name__ not in SUPPORTED_BATCH_SOURCES
            and source.to_proto().type != DataSourceProto.SourceType.CUSTOM_SOURCE
        ):
            raise ValueError(
                f"Batch feature views need a batch source, expected one of {SUPPORTED_BATCH_SOURCES} "
                f"or CUSTOM_SOURCE, got {type(source).__name__}: {source.name} instead "
            )

        self.mode = mode
        self.udf = udf
        self.udf_string = udf_string
        self.feature_transformation = (
            feature_transformation or self.get_feature_transformation()
        )
        self.batch_engine = batch_engine
        self.aggregations = aggregations or []

        super().__init__(
            name=name,
            entities=entities,
            ttl=ttl,
            tags=tags,
            online=online,
            offline=offline,
            description=description,
            owner=owner,
            schema=schema,
            source=source,  # type: ignore[arg-type]
            sink_source=sink_source,
        )

    def get_feature_transformation(self) -> Optional[Transformation]:
        if not self.udf:
            return None
        if self.mode in (
            TransformationMode.PANDAS,
            TransformationMode.PYTHON,
            TransformationMode.SQL,
        ) or self.mode in ("pandas", "python", "sql"):
            return Transformation(
                mode=self.mode, udf=self.udf, udf_string=self.udf_string or ""
            )
        else:
            raise ValueError(
                f"Unsupported transformation mode: {self.mode} for StreamFeatureView"
            )


def batch_feature_view(
    *,
    name: Optional[str] = None,
    mode: Union[TransformationMode, str] = TransformationMode.PYTHON,
    entities: Optional[List[str]] = None,
    ttl: Optional[timedelta] = None,
    source: Optional[DataSource] = None,
    tags: Optional[Dict[str, str]] = None,
    online: bool = True,
    offline: bool = True,
    description: str = "",
    owner: str = "",
    schema: Optional[List[Field]] = None,
):
    """
    Creates a BatchFeatureView object with the given user-defined function (UDF) as the transformation.
    Please make sure that the udf contains all non-built in imports within the function to ensure that the execution
    of a deserialized function does not miss imports.
    """

    def mainify(obj):
        # Needed to allow dill to properly serialize the udf. Otherwise, clients will need to have a file with the same
        # name as the original file defining the sfv.
        if obj.__module__ != "__main__":
            obj.__module__ = "__main__"

    def decorator(user_function):
        udf_string = dill.source.getsource(user_function)
        mainify(user_function)
        batch_feature_view_obj = BatchFeatureView(
            name=name or user_function.__name__,
            mode=mode,
            entities=entities,
            ttl=ttl,
            source=source,
            tags=tags,
            online=online,
            offline=offline,
            description=description,
            owner=owner,
            schema=schema,
            udf=user_function,
            udf_string=udf_string,
        )
        functools.update_wrapper(wrapper=batch_feature_view_obj, wrapped=user_function)
        return batch_feature_view_obj

    return decorator



import logging

import grpc

from feast.errors import FeastError
from feast.permissions.auth.auth_type import AuthType
from feast.permissions.auth_model import AuthConfig
from feast.permissions.client.client_auth_token import get_auth_token

logger = logging.getLogger(__name__)


class GrpcClientAuthHeaderInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.StreamStreamClientInterceptor,
):
    def __init__(self, auth_config: AuthConfig):
        self._auth_config = auth_config

    def intercept_unary_unary(
        self, continuation, client_call_details, request_iterator
    ):
        return self._handle_call(continuation, client_call_details, request_iterator)

    def intercept_unary_stream(
        self, continuation, client_call_details, request_iterator
    ):
        return self._handle_call(continuation, client_call_details, request_iterator)

    def intercept_stream_unary(
        self, continuation, client_call_details, request_iterator
    ):
        return self._handle_call(continuation, client_call_details, request_iterator)

    def intercept_stream_stream(
        self, continuation, client_call_details, request_iterator
    ):
        return self._handle_call(continuation, client_call_details, request_iterator)

    def _handle_call(self, continuation, client_call_details, request_iterator):
        if self._auth_config.type != AuthType.NONE.value:
            client_call_details = self._append_auth_header_metadata(client_call_details)
        result = continuation(client_call_details, request_iterator)
        if result.exception() is not None:
            mapped_error = FeastError.from_error_detail(result.exception().details())
            if mapped_error is not None:
                raise mapped_error
        return result

    def _append_auth_header_metadata(self, client_call_details):
        logger.debug(
            "Intercepted the grpc api method call to inject Authorization header "
        )
        metadata = client_call_details.metadata or []
        access_token = get_auth_token(self._auth_config)
        metadata.append((b"authorization", b"Bearer " + access_token.encode("utf-8")))
        client_call_details = client_call_details._replace(metadata=metadata)
        return client_call_details



import re
from abc import ABC

from starlette.authentication import (
    AuthenticationError,
)


class TokenExtractor(ABC):
    """
    A class to extract the authorization token from a user request.
    """

    def extract_access_token(self, **kwargs) -> str:
        """
        Extract the authorization token from a user request.

        The actual implementation has to specify what arguments have to be defined in the kwywork args `kwargs`

        Returns:
            The extracted access token.
        """
        raise NotImplementedError()

    def _extract_bearer_token(self, auth_header: str) -> str:
        """
        Extract the bearer token from the authorization header value.

        Args:
            auth_header: The full value of the authorization header.

        Returns:
            str: The token value, without the `Bearer` part.

        Raises:
            AuthenticationError if the authorization token does not match the `Bearer` scheme.
        """
        pattern = r"(?i)Bearer .+"
        if not bool(re.match(pattern, auth_header)):
            raise AuthenticationError(f"Expected Bearer schema, found {auth_header}")
        _, access_token = auth_header.split()
        return access_token


class NoAuthTokenExtractor(TokenExtractor):
    """
    A `TokenExtractor` always returning an empty token
    """

    def extract_access_token(self, **kwargs) -> str:
        return ""



from functools import cache

import clickhouse_connect
from clickhouse_connect.driver import Client

from feast.infra.utils.clickhouse.clickhouse_config import ClickhouseConfig


@cache
def get_client(config: ClickhouseConfig) -> Client:
    client = clickhouse_connect.get_client(
        host=config.host,
        port=config.port,
        user=config.user,
        password=config.password,
        database=config.database,
    )
    return client



import hashlib
import logging
from concurrent import futures
from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Set, Tuple

import google
from google.cloud import bigtable
from google.cloud.bigtable import row_filters
from pydantic import StrictStr

from feast import Entity, FeatureView, utils
from feast.feature_view import DUMMY_ENTITY_NAME
from feast.infra.online_stores.helpers import compute_entity_id
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel, RepoConfig

logger = logging.getLogger(__name__)

# Number of mutations per Bigtable write operation we're aiming for. The official max is
# 100K; we're being conservative.
MUTATIONS_PER_OP = 50_000
# The Bigtable client library limits the connection pool size to 10. This imposes a
# limitation to the concurrency we can get using a thread pool in each worker.
BIGTABLE_CLIENT_CONNECTION_POOL_SIZE = 10


class BigtableOnlineStoreConfig(FeastConfigBaseModel):
    """Online store config for GCP Bigtable"""

    type: Literal["bigtable"] = "bigtable"
    """Online store typee selector"""

    project_id: Optional[StrictStr] = None
    """(optional) GCP Project ID"""

    instance: StrictStr
    """The Bigtable instance's ID"""

    max_versions: int = 2
    """The number of historical versions of data that will be kept around."""


class BigtableOnlineStore(OnlineStore):
    _client: Optional[bigtable.Client] = None

    feature_column_family: str = "features"

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        # Potential performance improvement opportunity described in
        # https://github.com/feast-dev/feast/issues/3259
        feature_view = table
        bt_table_name = self._get_table_name(config=config, feature_view=feature_view)

        client = self._get_client(online_config=config.online_store)
        bt_instance = client.instance(instance_id=config.online_store.instance)
        bt_table = bt_instance.table(bt_table_name)
        row_keys = [
            self._compute_row_key(
                entity_key=entity_key,
                feature_view_name=feature_view.name,
                config=config,
            )
            for entity_key in entity_keys
        ]

        row_set = bigtable.row_set.RowSet()
        for row_key in row_keys:
            row_set.add_row_key(row_key)
        rows = bt_table.read_rows(
            row_set=row_set,
            filter_=(
                row_filters.ColumnQualifierRegexFilter(
                    f"^({'|'.join(requested_features)}|event_ts)$".encode()
                )
                if requested_features
                else None
            ),
        )

        # The BigTable client library only returns rows for keys that are found. This
        # means that it's our responsibility to match the returned rows to the original
        # `row_keys` and make sure that we're returning a list of the same length as
        # `entity_keys`.
        bt_rows_dict: Dict[bytes, bigtable.row.PartialRowData] = {
            row.row_key: row for row in rows
        }
        return [self._process_bt_row(bt_rows_dict.get(row_key)) for row_key in row_keys]

    def _process_bt_row(
        self, row: Optional[bigtable.row.PartialRowData]
    ) -> Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]:
        res = {}

        if row is None:
            return (None, None)

        row_values = row.cells[self.feature_column_family]
        event_ts = datetime.fromisoformat(row_values.pop(b"event_ts")[0].value.decode())
        for feature_name, feature_values in row_values.items():
            # We only want to retrieve the latest value for each feature
            feature_value = feature_values[0]
            val = ValueProto()
            val.ParseFromString(feature_value.value)
            res[feature_name.decode()] = val

        return (event_ts, res)

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        feature_view = table
        bt_table_name = self._get_table_name(config=config, feature_view=feature_view)

        client = self._get_client(online_config=config.online_store)
        bt_instance = client.instance(instance_id=config.online_store.instance)
        bt_table = bt_instance.table(bt_table_name)

        # `columns_per_row` is used to calculate the number of rows we are allowed to
        # mutate in one request.
        columns_per_row = len(feature_view.features) + 1  # extra for event timestamp
        rows_per_write = MUTATIONS_PER_OP // columns_per_row

        with futures.ThreadPoolExecutor(
            max_workers=BIGTABLE_CLIENT_CONNECTION_POOL_SIZE
        ) as executor:
            fs = []
            while data:
                rows_to_write, data = data[:rows_per_write], data[rows_per_write:]
                fs.append(
                    executor.submit(
                        self._write_rows_to_bt,
                        rows_to_write=rows_to_write,
                        bt_table=bt_table,
                        feature_view_name=feature_view.name,
                        config=config,
                        progress=progress,
                    )
                )
            done_tasks, not_done_tasks = futures.wait(fs)
            for task in done_tasks:
                # If a task raised an exception, this will raise it here as well
                task.result()
            if not_done_tasks:
                raise RuntimeError(
                    f"Not all batches were written to Bigtable: {not_done_tasks}"
                )

    def _write_rows_to_bt(
        self,
        rows_to_write: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        bt_table: bigtable.table.Table,
        feature_view_name: str,
        config: RepoConfig,
        progress: Optional[Callable[[int], Any]],
    ):
        rows = []
        for entity_key, features, timestamp, created_ts in rows_to_write:
            bt_row = bt_table.direct_row(
                self._compute_row_key(
                    entity_key=entity_key,
                    feature_view_name=feature_view_name,
                    config=config,
                )
            )

            for feature_name, feature_value in features.items():
                bt_row.set_cell(
                    self.feature_column_family,
                    feature_name.encode(),
                    feature_value.SerializeToString(),
                )
            bt_row.set_cell(
                self.feature_column_family,
                b"event_ts",
                utils.make_tzaware(timestamp).isoformat().encode(),
            )
            rows.append(bt_row)
        bt_table.mutate_rows(rows)

        if progress:
            progress(len(rows))

    def _compute_row_key(
        self, entity_key: EntityKeyProto, feature_view_name: str, config: RepoConfig
    ) -> bytes:
        entity_id = compute_entity_id(
            entity_key,
            entity_key_serialization_version=config.entity_key_serialization_version,
        )
        # Even though `entity_id` uniquely identifies an entity, we use the same table
        # for multiple feature_views with the same set of entities. To uniquely identify
        # the row for a feature_view, we suffix the name of the feature_view itself.
        # This also ensures that features for entities from various feature_views are
        # colocated.
        return f"{entity_id}#{feature_view_name}".encode()

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        """Creates the appropriate tables and column family in Bigtable.

        We use a dedicated table for each entity combination. For example, if a
        FeatureView uses the entities `shop` and `customer`, the resulting table would
        be called `customer-shop` (entities are sorted lexicographically first).
        """
        online_config = config.online_store
        assert isinstance(online_config, BigtableOnlineStoreConfig)
        client = self._get_client(online_config, admin=True)
        bt_instance = client.instance(instance_id=online_config.instance)
        max_versions_gc_rule = bigtable.column_family.MaxVersionsGCRule(
            online_config.max_versions
        )
        # The word "table" in the arguments refers to feature_views (this is for legacy
        # reasons). To reduce confusion with bigtable tables, we use alternate variable
        # names
        feature_views_to_keep = tables_to_keep
        feature_views_to_delete = tables_to_delete
        # Multiple feature views can share the same tables. So just because a feature
        # view has been deleted does not mean that we can just delete the table. We map
        # feature views to Bigtable table names and figure out which ones to create
        # and/or delete.
        bt_tables_to_keep: Set[str] = {
            self._get_table_name(config=config, feature_view=feature_view)
            for feature_view in feature_views_to_keep
        }
        bt_tables_to_delete: Set[str] = {
            self._get_table_name(config=config, feature_view=feature_view)
            for feature_view in feature_views_to_delete
        } - bt_tables_to_keep  # we don't delete a table if it's in `bt_tables_to_keep`

        for bt_table_name in bt_tables_to_keep:
            bt_table = bt_instance.table(bt_table_name)
            if not bt_table.exists():
                logger.info(f"Creating table `{bt_table_name}` in Bigtable")
                bt_table.create()
            else:
                logger.info(f"Table {bt_table_name} already exists in Bigtable")

            if self.feature_column_family not in bt_table.list_column_families():
                bt_table.column_family(
                    self.feature_column_family, gc_rule=max_versions_gc_rule
                ).create()

        for bt_table_name in bt_tables_to_delete:
            bt_table = bt_instance.table(bt_table_name)
            logger.info(f"Deleting table {bt_table_name} in Bigtable")
            bt_table.delete()

    @staticmethod
    def _get_table_name(config: RepoConfig, feature_view: FeatureView) -> str:
        entities_part = (
            "-".join(sorted(feature_view.entities))
            if feature_view.entities
            else DUMMY_ENTITY_NAME
        )
        BIGTABLE_TABLE_MAX_LENGTH = 50
        ENTITIES_PART_MAX_LENGTH = 24
        # Bigtable limits table names to 50 characters. We'll limit the max size of of
        # the `entities_part` and if that's not enough, we'll just hash the
        # entities_part. The remaining length is dedicated to the project name. This
        # allows multiple projects to coexist in the same bigtable instance. This also
        # allows multiple integration test executions to run simultaneously without
        # conflicts.
        if len(entities_part) > ENTITIES_PART_MAX_LENGTH:
            entities_part = hashlib.md5(entities_part.encode()).hexdigest()[
                :ENTITIES_PART_MAX_LENGTH
            ]
        remaining_length = BIGTABLE_TABLE_MAX_LENGTH - len(entities_part) - 1
        if len(config.project) > remaining_length:
            HUMAN_READABLE_PART_LENGTH = 10
            HASH_PART_LENGTH = remaining_length - HUMAN_READABLE_PART_LENGTH - 1
            project_part = (
                config.project[:HUMAN_READABLE_PART_LENGTH]
                + "_"
                + hashlib.md5(config.project.encode()).hexdigest()[:HASH_PART_LENGTH]
            )
        else:
            project_part = config.project
        return f"{project_part}.{entities_part}"

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        # Because of historical reasons, Feast calls them tables. We use this alias for
        # readability.
        feature_views = tables

        bt_tables = {
            self._get_table_name(config=config, feature_view=fv) for fv in feature_views
        }

        online_config = config.online_store
        assert isinstance(online_config, BigtableOnlineStoreConfig)
        client = self._get_client(online_config, admin=True)
        bt_instance = client.instance(instance_id=online_config.instance)
        for table_name in bt_tables:
            try:
                logger.info(f"Deleting Bigtable table `{table_name}`")
                bt_instance.table(table_name).delete()
            except google.api_core.exceptions.NotFound:
                logger.warning(
                    f"Table `{table_name}` was not found. Skipping deletion."
                )

    def _get_client(
        self, online_config: BigtableOnlineStoreConfig, admin: bool = False
    ):
        if self._client is None:
            self._client = bigtable.Client(
                project=online_config.project_id, admin=admin
            )
        return self._client



# Copyright 2021 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from collections import defaultdict
from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple

import requests
from pydantic import StrictStr

from feast import Entity, FeatureView, RepoConfig
from feast.infra.online_stores.helpers import _to_naive_utc
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel
from feast.rest_error_handler import rest_error_handling_decorator
from feast.type_map import (
    feast_value_type_to_python_type,
    python_values_to_proto_values,
)
from feast.utils import _get_feature_view_vector_field_metadata
from feast.value_type import ValueType

logger = logging.getLogger(__name__)


class RemoteOnlineStoreConfig(FeastConfigBaseModel):
    """Remote Online store config for remote online store"""

    type: Literal["remote"] = "remote"
    """Online store type selector"""

    path: StrictStr = "http://localhost:6566"
    """ str: Path to metadata store.
    If type is 'remote', then this is a URL for registry server """

    cert: StrictStr = ""
    """ str: Path to the public certificate when the online server starts in TLS(SSL) mode. This may be needed if the online server started with a self-signed certificate, typically this file ends with `*.crt`, `*.cer`, or `*.pem`.
    If type is 'remote', then this configuration is needed to connect to remote online server in TLS mode. """


class RemoteOnlineStore(OnlineStore):
    """
    remote online store implementation wrapper to communicate with feast online server.
    """

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        """
        Writes a batch of feature rows to the remote online store via the remote API.
        """
        assert isinstance(config.online_store, RemoteOnlineStoreConfig)
        config.online_store.__class__ = RemoteOnlineStoreConfig

        columnar_data: Dict[str, List[Any]] = defaultdict(list)

        # Iterate through each row to populate columnar data directly
        for entity_key_proto, feature_values_proto, event_ts, created_ts in data:
            # Populate entity key values
            for join_key, entity_value_proto in zip(
                entity_key_proto.join_keys, entity_key_proto.entity_values
            ):
                columnar_data[join_key].append(
                    feast_value_type_to_python_type(entity_value_proto)
                )

            # Populate feature values
            for feature_name, feature_value_proto in feature_values_proto.items():
                columnar_data[feature_name].append(
                    feast_value_type_to_python_type(feature_value_proto)
                )

            # Populate timestamps
            columnar_data["event_timestamp"].append(_to_naive_utc(event_ts).isoformat())
            columnar_data["created"].append(
                _to_naive_utc(created_ts).isoformat() if created_ts else None
            )

        req_body = {
            "feature_view_name": table.name,
            "df": columnar_data,
            "allow_registry_cache": False,
        }

        response = post_remote_online_write(config=config, req_body=req_body)

        if response.status_code != 200:
            error_msg = f"Unable to write online store data using feature server API. Error_code={response.status_code}, error_message={response.text}"
            logger.error(error_msg)
            raise RuntimeError(error_msg)

        if progress:
            data_length = len(data)
            logger.info(
                f"Writing {data_length} rows to the remote store for feature view {table.name}."
            )
            progress(data_length)

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        assert isinstance(config.online_store, RemoteOnlineStoreConfig)
        config.online_store.__class__ = RemoteOnlineStoreConfig

        req_body = self._construct_online_read_api_json_request(
            entity_keys, table, requested_features
        )
        response = get_remote_online_features(config=config, req_body=req_body)
        if response.status_code == 200:
            logger.debug("Able to retrieve the online features from feature server.")
            response_json = json.loads(response.text)
            event_ts = self._get_event_ts(response_json)
            # Iterating over results and converting the API results in column format to row format.
            result_tuples: List[
                Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]
            ] = []
            for feature_value_index in range(len(entity_keys)):
                feature_values_dict: Dict[str, ValueProto] = dict()
                for index, feature_name in enumerate(
                    response_json["metadata"]["feature_names"]
                ):
                    if (
                        requested_features is not None
                        and feature_name in requested_features
                    ):
                        if (
                            response_json["results"][index]["statuses"][
                                feature_value_index
                            ]
                            == "PRESENT"
                        ):
                            message = python_values_to_proto_values(
                                [
                                    response_json["results"][index]["values"][
                                        feature_value_index
                                    ]
                                ],
                                ValueType.UNKNOWN,
                            )
                            feature_values_dict[feature_name] = message[0]
                        else:
                            feature_values_dict[feature_name] = ValueProto()
                result_tuples.append((event_ts, feature_values_dict))
            return result_tuples
        else:
            error_msg = f"Unable to retrieve the online store data using feature server API. Error_code={response.status_code}, error_message={response.text}"
            logger.error(error_msg)
            raise RuntimeError(error_msg)

    def retrieve_online_documents(
        self,
        config: RepoConfig,
        table: FeatureView,
        requested_features: Optional[List[str]],
        embedding: Optional[List[float]],
        top_k: int,
        distance_metric: Optional[str] = "L2",
    ) -> List[
        Tuple[
            Optional[datetime],
            Optional[EntityKeyProto],
            Optional[ValueProto],
            Optional[ValueProto],
            Optional[ValueProto],
        ]
    ]:
        assert isinstance(config.online_store, RemoteOnlineStoreConfig)
        config.online_store.__class__ = RemoteOnlineStoreConfig

        req_body = self._construct_online_documents_api_json_request(
            table, requested_features, embedding, top_k, distance_metric
        )
        response = get_remote_online_documents(config=config, req_body=req_body)
        if response.status_code == 200:
            logger.debug("Able to retrieve the online documents from feature server.")
            response_json = json.loads(response.text)
            event_ts: Optional[datetime] = self._get_event_ts(response_json)

            # Create feature name to index mapping for efficient lookup
            feature_name_to_index = {
                name: idx
                for idx, name in enumerate(response_json["metadata"]["feature_names"])
            }

            vector_field_metadata = _get_feature_view_vector_field_metadata(table)

            # Process each result row
            num_results = len(response_json["results"][0]["values"])
            result_tuples = []

            for row_idx in range(num_results):
                # Extract values using helper methods
                feature_val = self._extract_requested_feature_value(
                    response_json, feature_name_to_index, requested_features, row_idx
                )
                vector_value = self._extract_vector_field_value(
                    response_json, feature_name_to_index, vector_field_metadata, row_idx
                )
                distance_val = self._extract_distance_value(
                    response_json, feature_name_to_index, "distance", row_idx
                )
                entity_key_proto = self._construct_entity_key_from_response(
                    response_json, row_idx, feature_name_to_index, table
                )

                result_tuples.append(
                    (
                        event_ts,
                        entity_key_proto,
                        feature_val,
                        vector_value,
                        distance_val,
                    )
                )

            return result_tuples
        else:
            error_msg = f"Unable to retrieve the online documents using feature server API. Error_code={response.status_code}, error_message={response.text}"
            logger.error(error_msg)
            raise RuntimeError(error_msg)

    def retrieve_online_documents_v2(
        self,
        config: RepoConfig,
        table: FeatureView,
        requested_features: Optional[List[str]],
        embedding: Optional[List[float]],
        top_k: int,
        distance_metric: Optional[str] = None,
        query_string: Optional[str] = None,
    ) -> List[
        Tuple[
            Optional[datetime],
            Optional[EntityKeyProto],
            Optional[Dict[str, ValueProto]],
        ]
    ]:
        assert isinstance(config.online_store, RemoteOnlineStoreConfig)
        config.online_store.__class__ = RemoteOnlineStoreConfig

        req_body = self._construct_online_documents_v2_api_json_request(
            table,
            requested_features,
            embedding,
            top_k,
            distance_metric,
            query_string,
            api_version=2,
        )
        response = get_remote_online_documents(config=config, req_body=req_body)
        if response.status_code == 200:
            logger.debug("Able to retrieve the online documents from feature server.")
            response_json = json.loads(response.text)
            event_ts: Optional[datetime] = self._get_event_ts(response_json)

            # Create feature name to index mapping for efficient lookup
            feature_name_to_index = {
                name: idx
                for idx, name in enumerate(response_json["metadata"]["feature_names"])
            }

            # Process each result row
            num_results = (
                len(response_json["results"][0]["values"])
                if response_json["results"]
                else 0
            )
            result_tuples = []

            for row_idx in range(num_results):
                # Build feature values dictionary for requested features
                feature_values_dict = {}

                if requested_features:
                    for feature_name in requested_features:
                        if feature_name in feature_name_to_index:
                            feature_idx = feature_name_to_index[feature_name]
                            if self._is_feature_present(
                                response_json, feature_idx, row_idx
                            ):
                                feature_values_dict[feature_name] = (
                                    self._extract_feature_value(
                                        response_json, feature_idx, row_idx
                                    )
                                )
                            else:
                                feature_values_dict[feature_name] = ValueProto()

                # Construct entity key proto using existing helper method
                entity_key_proto = self._construct_entity_key_from_response(
                    response_json, row_idx, feature_name_to_index, table
                )

                result_tuples.append(
                    (
                        event_ts,
                        entity_key_proto,
                        feature_values_dict if feature_values_dict else None,
                    )
                )

            return result_tuples
        else:
            error_msg = f"Unable to retrieve the online documents using feature server API. Error_code={response.status_code}, error_message={response.text}"
            logger.error(error_msg)
            raise RuntimeError(error_msg)

    def _extract_requested_feature_value(
        self,
        response_json: dict,
        feature_name_to_index: dict,
        requested_features: Optional[List[str]],
        row_idx: int,
    ) -> Optional[ValueProto]:
        """Extract the first available requested feature value."""
        if not requested_features:
            return ValueProto()

        for feature_name in requested_features:
            if feature_name in feature_name_to_index:
                feature_idx = feature_name_to_index[feature_name]
                if self._is_feature_present(response_json, feature_idx, row_idx):
                    return self._extract_feature_value(
                        response_json, feature_idx, row_idx
                    )

        return ValueProto()

    def _extract_vector_field_value(
        self,
        response_json: dict,
        feature_name_to_index: dict,
        vector_field_metadata,
        row_idx: int,
    ) -> Optional[ValueProto]:
        """Extract vector field value from response."""
        if (
            not vector_field_metadata
            or vector_field_metadata.name not in feature_name_to_index
        ):
            return ValueProto()

        vector_feature_idx = feature_name_to_index[vector_field_metadata.name]
        if self._is_feature_present(response_json, vector_feature_idx, row_idx):
            return self._extract_feature_value(
                response_json, vector_feature_idx, row_idx
            )

        return ValueProto()

    def _extract_distance_value(
        self,
        response_json: dict,
        feature_name_to_index: dict,
        distance_feature_name: str,
        row_idx: int,
    ) -> Optional[ValueProto]:
        """Extract distance/score value from response."""
        if not distance_feature_name:
            return ValueProto()

        distance_feature_idx = feature_name_to_index[distance_feature_name]
        if self._is_feature_present(response_json, distance_feature_idx, row_idx):
            distance_value = response_json["results"][distance_feature_idx]["values"][
                row_idx
            ]
            distance_val = ValueProto()
            distance_val.float_val = float(distance_value)
            return distance_val

        return ValueProto()

    def _is_feature_present(
        self, response_json: dict, feature_idx: int, row_idx: int
    ) -> bool:
        """Check if a feature is present in the response."""
        return response_json["results"][feature_idx]["statuses"][row_idx] == "PRESENT"

    def _construct_online_read_api_json_request(
        self,
        entity_keys: List[EntityKeyProto],
        table: FeatureView,
        requested_features: Optional[List[str]] = None,
    ) -> str:
        api_requested_features = []
        if requested_features is not None:
            for requested_feature in requested_features:
                api_requested_features.append(f"{table.name}:{requested_feature}")

        entity_values = []
        entity_key = ""
        for row in entity_keys:
            entity_key = row.join_keys[0]
            entity_values.append(
                getattr(row.entity_values[0], row.entity_values[0].WhichOneof("val"))
            )

        req_body = json.dumps(
            {
                "features": api_requested_features,
                "entities": {entity_key: entity_values},
            }
        )
        return req_body

    def _construct_online_documents_api_json_request(
        self,
        table: FeatureView,
        requested_features: Optional[List[str]] = None,
        embedding: Optional[List[float]] = None,
        top_k: Optional[int] = None,
        distance_metric: Optional[str] = "L2",
    ) -> str:
        api_requested_features = []
        if requested_features is not None:
            for requested_feature in requested_features:
                api_requested_features.append(f"{table.name}:{requested_feature}")

        req_body = json.dumps(
            {
                "features": api_requested_features,
                "query": embedding,
                "top_k": top_k,
                "distance_metric": distance_metric,
            }
        )
        return req_body

    def _construct_online_documents_v2_api_json_request(
        self,
        table: FeatureView,
        requested_features: Optional[List[str]],
        embedding: Optional[List[float]],
        top_k: int,
        distance_metric: Optional[str] = None,
        query_string: Optional[str] = None,
        api_version: Optional[int] = 2,
    ) -> str:
        api_requested_features = []
        if requested_features is not None:
            for requested_feature in requested_features:
                api_requested_features.append(f"{table.name}:{requested_feature}")

        req_body = json.dumps(
            {
                "features": api_requested_features,
                "query": embedding,
                "top_k": top_k,
                "distance_metric": distance_metric,
                "query_string": query_string,
                "api_version": api_version,
            }
        )
        return req_body

    def _get_event_ts(self, response_json) -> datetime:
        event_ts = ""
        if len(response_json["results"]) > 1:
            event_ts = response_json["results"][1]["event_timestamps"][0]
        return datetime.fromisoformat(event_ts.replace("Z", "+00:00"))

    def _construct_entity_key_from_response(
        self,
        response_json: dict,
        row_idx: int,
        feature_name_to_index: dict,
        table: FeatureView,
    ) -> Optional[EntityKeyProto]:
        """Construct EntityKeyProto from response data."""
        # Use the feature view's join_keys to identify entity fields
        entity_fields = [
            join_key
            for join_key in table.join_keys
            if join_key in feature_name_to_index
        ]

        if not entity_fields:
            return None

        entity_key_proto = EntityKeyProto()
        entity_key_proto.join_keys.extend(entity_fields)

        for entity_field in entity_fields:
            if entity_field in feature_name_to_index:
                feature_idx = feature_name_to_index[entity_field]
                if self._is_feature_present(response_json, feature_idx, row_idx):
                    entity_value = self._extract_feature_value(
                        response_json, feature_idx, row_idx
                    )
                    entity_key_proto.entity_values.append(entity_value)

        return entity_key_proto if entity_key_proto.entity_values else None

    def _extract_feature_value(
        self, response_json: dict, feature_idx: int, row_idx: int
    ) -> ValueProto:
        """Extract and convert a feature value to ValueProto."""
        raw_value = response_json["results"][feature_idx]["values"][row_idx]
        if raw_value is None:
            return ValueProto()
        proto_values = python_values_to_proto_values([raw_value])
        return proto_values[0]

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        pass

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        pass


@rest_error_handling_decorator
def get_remote_online_features(
    session: requests.Session, config: RepoConfig, req_body: str
) -> requests.Response:
    if config.online_store.cert:
        return session.post(
            f"{config.online_store.path}/get-online-features",
            data=req_body,
            verify=config.online_store.cert,
        )
    else:
        return session.post(
            f"{config.online_store.path}/get-online-features", data=req_body
        )


@rest_error_handling_decorator
def get_remote_online_documents(
    session: requests.Session, config: RepoConfig, req_body: str
) -> requests.Response:
    if config.online_store.cert:
        return session.post(
            f"{config.online_store.path}/retrieve-online-documents",
            data=req_body,
            verify=config.online_store.cert,
        )
    else:
        return session.post(
            f"{config.online_store.path}/retrieve-online-documents", data=req_body
        )


@rest_error_handling_decorator
def post_remote_online_write(
    session: requests.Session, config: RepoConfig, req_body: dict
) -> requests.Response:
    url = f"{config.online_store.path}/write-to-online-store"
    if config.online_store.cert:
        return session.post(url, json=req_body, verify=config.online_store.cert)
    else:
        return session.post(url, json=req_body)



from __future__ import absolute_import

from collections import defaultdict
from datetime import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple

import singlestoredb
from pydantic import StrictStr
from singlestoredb.connection import Connection, Cursor
from singlestoredb.exceptions import InterfaceError

from feast import Entity, FeatureView, RepoConfig
from feast.infra.key_encoding_utils import serialize_entity_key
from feast.infra.online_stores.helpers import _to_naive_utc
from feast.infra.online_stores.online_store import OnlineStore
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel


class SingleStoreOnlineStoreConfig(FeastConfigBaseModel):
    """
    Configuration for the SingleStore online store.
    NOTE: The class *must* end with the `OnlineStoreConfig` suffix.
    """

    type: Literal["singlestore"] = "singlestore"

    host: Optional[StrictStr] = None
    user: Optional[StrictStr] = None
    password: Optional[StrictStr] = None
    database: Optional[StrictStr] = None
    port: Optional[int] = None


class SingleStoreOnlineStore(OnlineStore):
    """
    An online store implementation that uses SingleStore.
    NOTE: The class *must* end with the `OnlineStore` suffix.
    """

    _conn: Optional[Connection] = None

    def _init_conn(self, config: RepoConfig) -> Connection:
        online_store_config = config.online_store
        assert isinstance(online_store_config, SingleStoreOnlineStoreConfig)
        return singlestoredb.connect(
            host=online_store_config.host or "127.0.0.1",
            user=online_store_config.user or "test",
            password=online_store_config.password or "test",
            database=online_store_config.database or "feast",
            port=online_store_config.port or 3306,
            conn_attrs={"_connector_name": "SingleStore Feast Online Store"},
            autocommit=True,
        )

    def _get_cursor(self, config: RepoConfig) -> Any:
        # This will try to reconnect also.
        # In case it fails, we will have to create a new connection.
        if not self._conn:
            self._conn = self._init_conn(config)
        try:
            self._conn.ping(reconnect=True)
        except InterfaceError:
            self._conn = self._init_conn(config)
        return self._conn.cursor()

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        project = config.project
        with self._get_cursor(config) as cur:
            insert_values = []
            for entity_key, values, timestamp, created_ts in data:
                entity_key_bin = serialize_entity_key(
                    entity_key,
                    entity_key_serialization_version=3,
                ).hex()
                timestamp = _to_naive_utc(timestamp)
                if created_ts is not None:
                    created_ts = _to_naive_utc(created_ts)

                for feature_name, val in values.items():
                    insert_values.append(
                        (
                            entity_key_bin,
                            feature_name,
                            val.SerializeToString(),
                            timestamp,
                            created_ts,
                        )
                    )
            # Control the batch so that we can update the progress
            batch_size = 50000
            for i in range(0, len(insert_values), batch_size):
                current_batch = insert_values[i : i + batch_size]
                cur.executemany(
                    f"""
                    INSERT INTO {_table_id(project, table)}
                    (entity_key, feature_name, value, event_ts, created_ts)
                    values (%s, %s, %s, %s, %s)
                    ON DUPLICATE KEY UPDATE
                    value = VALUES(value),
                    event_ts = VALUES(event_ts),
                    created_ts = VALUES(created_ts);
                    """,
                    current_batch,
                )
                if progress:
                    progress(len(current_batch))

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        project = config.project
        result: List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]] = []
        with self._get_cursor(config) as cur:
            keys = []
            for entity_key in entity_keys:
                keys.append(
                    serialize_entity_key(
                        entity_key,
                        entity_key_serialization_version=3,
                    ).hex()
                )

            if not requested_features:
                entity_key_placeholders = ",".join(["%s" for _ in keys])
                cur.execute(
                    f"""
                    SELECT entity_key, feature_name, value, event_ts FROM {_table_id(project, table)}
                    WHERE entity_key IN ({entity_key_placeholders})
                    ORDER BY event_ts;
                    """,
                    tuple(keys),
                )
            else:
                entity_key_placeholders = ",".join(["%s" for _ in keys])
                requested_features_placeholders = ",".join(
                    ["%s" for _ in requested_features]
                )
                cur.execute(
                    f"""
                    SELECT entity_key, feature_name, value, event_ts FROM {_table_id(project, table)}
                    WHERE entity_key IN ({entity_key_placeholders}) and feature_name IN ({requested_features_placeholders})
                    ORDER BY event_ts;
                    """,
                    tuple(keys + requested_features),
                )
            rows = cur.fetchall() or []

            # Since we don't know the order returned from MySQL we'll need
            # to construct a dict to be able to quickly look up the correct row
            # when we iterate through the keys since they are in the correct order
            values_dict = defaultdict(list)
            for row in rows:
                values_dict[row[0]].append(row[1:])

            for key in keys:
                if key in values_dict:
                    key_values = values_dict[key]
                    res = {}
                    res_ts: Optional[datetime] = None
                    for feature_name, value_bin, event_ts in key_values:
                        val = ValueProto()
                        val.ParseFromString(bytes(value_bin))
                        res[feature_name] = val
                        res_ts = event_ts
                    result.append((res_ts, res))
                else:
                    result.append((None, None))
            return result

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ) -> None:
        project = config.project
        with self._get_cursor(config) as cur:
            # We don't create any special state for the entities in this implementation.
            for table in tables_to_keep:
                cur.execute(
                    f"""CREATE TABLE IF NOT EXISTS {_table_id(project, table)} (entity_key VARCHAR(512),
                    feature_name VARCHAR(256),
                    value BLOB,
                    event_ts timestamp NULL DEFAULT NULL,
                    created_ts timestamp NULL DEFAULT NULL,
                    PRIMARY KEY(entity_key, feature_name),
                    INDEX {_table_id(project, table)}_ek (entity_key))"""
                )

            for table in tables_to_delete:
                _drop_table_and_index(cur, project, table)

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ) -> None:
        project = config.project
        with self._get_cursor(config) as cur:
            for table in tables:
                _drop_table_and_index(cur, project, table)


def _drop_table_and_index(cur: Cursor, project: str, table: FeatureView) -> None:
    table_name = _table_id(project, table)
    cur.execute(f"DROP INDEX {table_name}_ek ON {table_name};")
    cur.execute(f"DROP TABLE IF EXISTS {table_name}")


def _table_id(project: str, table: FeatureView) -> str:
    return f"{project}_{table.name}"



import logging
import uuid
from dataclasses import dataclass
from typing import List, Optional

import pandas as pd
import pyarrow as pa
import ray
from ray.data import Dataset

from feast import OnDemandFeatureView
from feast.dqm.errors import ValidationFailed
from feast.errors import SavedDatasetLocationAlreadyExists
from feast.infra.common.materialization_job import (
    MaterializationJob,
    MaterializationJobStatus,
)
from feast.infra.compute_engines.dag.context import ExecutionContext
from feast.infra.compute_engines.dag.model import DAGFormat
from feast.infra.compute_engines.dag.plan import ExecutionPlan
from feast.infra.compute_engines.dag.value import DAGValue
from feast.infra.offline_stores.file_source import SavedDatasetFileStorage
from feast.infra.offline_stores.offline_store import RetrievalJob, RetrievalMetadata
from feast.repo_config import RepoConfig
from feast.saved_dataset import SavedDatasetStorage

logger = logging.getLogger(__name__)


class RayDAGRetrievalJob(RetrievalJob):
    """
    Ray-based retrieval job that executes a DAG plan to retrieve historical features.
    """

    def __init__(
        self,
        plan: Optional[ExecutionPlan],
        context: Optional[ExecutionContext],
        config: RepoConfig,
        full_feature_names: bool,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
        feature_refs: Optional[List[str]] = None,
        metadata: Optional[RetrievalMetadata] = None,
        error: Optional[BaseException] = None,
    ):
        super().__init__()
        self._plan = plan
        self._context = context
        self._config = config
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = on_demand_feature_views or []
        self._feature_refs = feature_refs or []
        self._metadata = metadata
        self._error = error
        self._result_dataset: Optional[Dataset] = None
        self._result_df: Optional[pd.DataFrame] = None
        self._result_arrow: Optional[pa.Table] = None

    def error(self) -> Optional[BaseException]:
        """Return any error that occurred during job execution."""
        return self._error

    def _ensure_executed(self) -> DAGValue:
        """Ensure the execution plan has been executed."""
        if self._result_dataset is None and self._plan and self._context:
            try:
                result = self._plan.execute(self._context)
                if hasattr(result, "data") and isinstance(result.data, Dataset):
                    self._result_dataset = result.data
                else:
                    # If result is not a Ray Dataset, convert it
                    if isinstance(result.data, pd.DataFrame):
                        self._result_dataset = ray.data.from_pandas(result.data)
                    elif isinstance(result.data, pa.Table):
                        self._result_dataset = ray.data.from_arrow(result.data)
                    else:
                        raise ValueError(
                            f"Unsupported result type: {type(result.data)}"
                        )
                return result
            except Exception as e:
                self._error = e
                logger.error(f"Ray DAG execution failed: {e}")
                raise
        elif self._result_dataset is None:
            raise ValueError("No execution plan available or execution failed")

        # Return a mock DAGValue for compatibility
        return DAGValue(data=self._result_dataset, format=DAGFormat.RAY)

    def to_ray_dataset(self) -> Dataset:
        """Get the result as a Ray Dataset."""
        self._ensure_executed()
        assert self._result_dataset is not None, (
            "Dataset should not be None after execution"
        )
        return self._result_dataset

    def to_df(
        self,
        validation_reference=None,
        timeout: Optional[int] = None,
    ) -> pd.DataFrame:
        """Convert the result to a pandas DataFrame."""
        if self._result_df is None:
            if self.on_demand_feature_views:
                # Use parent implementation for ODFV processing
                logger.info(
                    f"Processing {len(self.on_demand_feature_views)} on-demand feature views"
                )
                self._result_df = super().to_df(
                    validation_reference=validation_reference, timeout=timeout
                )
            else:
                # Direct conversion from Ray Dataset
                self._ensure_executed()
                assert self._result_dataset is not None, (
                    "Dataset should not be None after execution"
                )
                self._result_df = self._result_dataset.to_pandas()

        # Handle validation if provided
        if validation_reference:
            try:
                validation_result = validation_reference.profile.validate(
                    self._result_df
                )
                if not validation_result.is_success:
                    raise ValidationFailed(validation_result)
            except ImportError:
                logger.warning("DQM profiler not available, skipping validation")
            except Exception as e:
                logger.error(f"Validation failed: {e}")
                raise ValueError(f"Data validation failed: {e}")

        return self._result_df

    def to_arrow(
        self,
        validation_reference=None,
        timeout: Optional[int] = None,
    ) -> pa.Table:
        """Convert the result to an Arrow Table."""
        if self._result_arrow is None:
            if self.on_demand_feature_views:
                # Use parent implementation for ODFV processing
                self._result_arrow = super().to_arrow(
                    validation_reference=validation_reference, timeout=timeout
                )
            else:
                # Direct conversion from Ray Dataset
                self._ensure_executed()
                assert self._result_dataset is not None, (
                    "Dataset should not be None after execution"
                )
                self._result_arrow = self._result_dataset.to_pandas().to_arrow()

        # Handle validation if provided
        if validation_reference:
            try:
                df = self._result_arrow.to_pandas()
                validation_result = validation_reference.profile.validate(df)
                if not validation_result.is_success:
                    raise ValidationFailed(validation_result)
            except ImportError:
                logger.warning("DQM profiler not available, skipping validation")
            except Exception as e:
                logger.error(f"Validation failed: {e}")
                raise ValueError(f"Data validation failed: {e}")

        return self._result_arrow

    def to_remote_storage(self) -> list[str]:
        """Write the result to remote storage."""
        if not self._config.batch_engine.staging_location:
            raise ValueError("Staging location must be set for remote storage")

        try:
            self._ensure_executed()
            assert self._result_dataset is not None, (
                "Dataset should not be None after execution"
            )
            output_uri = (
                f"{self._config.batch_engine.staging_location}/{str(uuid.uuid4())}"
            )
            self._result_dataset.write_parquet(output_uri)
            logger.debug(f"Wrote result to {output_uri}")
            return [output_uri]
        except Exception as e:
            raise RuntimeError(f"Failed to write to remote storage: {e}")

    def persist(
        self,
        storage: SavedDatasetStorage,
        allow_overwrite: bool = False,
        timeout: Optional[int] = None,
    ) -> str:
        """Persist the result to the specified storage."""
        if not isinstance(storage, SavedDatasetFileStorage):
            raise ValueError(
                f"Ray compute engine only supports SavedDatasetFileStorage, got {type(storage)}"
            )

        destination_path = storage.file_options.uri

        # Check if destination already exists
        if not destination_path.startswith(("s3://", "gs://", "hdfs://")):
            import os

            if not allow_overwrite and os.path.exists(destination_path):
                raise SavedDatasetLocationAlreadyExists(location=destination_path)
            os.makedirs(os.path.dirname(destination_path), exist_ok=True)

        try:
            self._ensure_executed()
            assert self._result_dataset is not None, (
                "Dataset should not be None after execution"
            )
            self._result_dataset.write_parquet(destination_path)
            return destination_path
        except Exception as e:
            raise RuntimeError(f"Failed to persist dataset to {destination_path}: {e}")

    def to_sql(self) -> str:
        """Generate SQL representation of the execution plan."""
        if self._plan and self._context:
            return self._plan.to_sql(self._context)
        raise NotImplementedError("SQL generation not available without execution plan")

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata

    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        """Internal method to get DataFrame (used by parent class)."""
        self._ensure_executed()
        assert self._result_dataset is not None, (
            "Dataset should not be None after execution"
        )
        return self._result_dataset.to_pandas()

    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pa.Table:
        """Internal method to get Arrow Table (used by parent class)."""
        self._ensure_executed()
        assert self._result_dataset is not None, (
            "Dataset should not be None after execution"
        )
        return self._result_dataset.to_pandas().to_arrow()


@dataclass
class RayMaterializationJob(MaterializationJob):
    """
    Ray-based materialization job that tracks the status of feature materialization.
    """

    def __init__(
        self,
        job_id: str,
        status: MaterializationJobStatus,
        result: Optional[DAGValue] = None,
        error: Optional[BaseException] = None,
    ):
        super().__init__()
        self._job_id = job_id
        self._status = status
        self._result = result
        self._error = error

    def job_id(self) -> str:
        return self._job_id

    def status(self) -> MaterializationJobStatus:
        return self._status

    def error(self) -> Optional[BaseException]:
        return self._error

    def should_be_retried(self) -> bool:
        """Ray jobs are generally not retried by default."""
        return False

    def url(self) -> Optional[str]:
        """Ray jobs don't have a specific URL."""
        return None

    def result(self) -> Optional[DAGValue]:
        """Get the result of the materialization job."""
        return self._result



from types import FunctionType
from typing import Any, Dict, Optional, cast

import dill
import pyarrow

from feast.field import Field, from_value_type
from feast.protos.feast.core.Transformation_pb2 import (
    UserDefinedFunctionV2 as UserDefinedFunctionProto,
)
from feast.transformation.base import Transformation
from feast.transformation.mode import TransformationMode
from feast.type_map import (
    python_type_to_feast_value_type,
)


class PythonTransformation(Transformation):
    udf: FunctionType

    def __new__(
        cls,
        udf: FunctionType,
        udf_string: str,
        singleton: bool = False,
        name: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        description: str = "",
        owner: str = "",
    ) -> "PythonTransformation":
        instance = super(PythonTransformation, cls).__new__(
            cls,
            mode=TransformationMode.PYTHON,
            singleton=singleton,
            udf=udf,
            udf_string=udf_string,
            name=name,
            tags=tags,
            description=description,
            owner=owner,
        )
        return cast(PythonTransformation, instance)

    def __init__(
        self,
        udf: FunctionType,
        udf_string: str,
        singleton: bool = False,
        name: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        description: str = "",
        owner: str = "",
        *args,
        **kwargs,
    ):
        """
        Creates a PythonTransformation object.

        Args:
            udf: The user-defined transformation function, which must take pandas
                dataframes as inputs.
            name: The name of the transformation.
            udf_string: The source code version of the UDF (for diffing and displaying in Web UI).
            tags: Metadata tags for the transformation.
            description: A description of the transformation.
            owner: The owner of the transformation.
        """
        super().__init__(
            mode=TransformationMode.PYTHON,
            udf=udf,
            name=name,
            udf_string=udf_string,
            tags=tags,
            description=description,
            owner=owner,
        )
        self.singleton = singleton

    def transform_arrow(
        self,
        pa_table: pyarrow.Table,
        features: list[Field],
    ) -> pyarrow.Table:
        return pyarrow.Table.from_pydict(self.udf(pa_table.to_pydict()))

    def transform(self, input_dict: dict) -> dict:
        # Ensuring that the inputs are included as well
        output_dict = self.udf.__call__(input_dict)
        return {**input_dict, **output_dict}

    def transform_singleton(self, input_dict: dict) -> dict:
        # This flattens the list of elements to extract the first one
        # in the case of a singleton element, it takes the value directly
        # in the case of a list of lists, it takes the first list
        input_dict = {k: v[0] for k, v in input_dict.items()}
        output_dict = self.udf.__call__(input_dict)
        return {**input_dict, **output_dict}

    def infer_features(
        self, random_input: dict[str, Any], singleton: Optional[bool] = False
    ) -> list[Field]:
        output_dict: dict[str, Any] = self.transform(random_input)

        fields = []
        for feature_name, feature_value in output_dict.items():
            if isinstance(feature_value, list):
                if len(feature_value) <= 0:
                    raise TypeError(
                        f"Failed to infer type for feature '{feature_name}' with value "
                        + f"'{feature_value}' since no items were returned by the UDF."
                    )
                inferred_value = feature_value[0]
                if singleton and isinstance(inferred_value, list):
                    # If we have a nested list like [[0.5, 0.5, ...]]
                    if len(inferred_value) > 0:
                        # Get the actual element type from the inner list
                        inferred_type = type(inferred_value[0])
                    else:
                        raise TypeError(
                            f"Failed to infer type for nested feature '{feature_name}' - inner list is empty"
                        )
                else:
                    # For non-nested lists or when singleton is False
                    inferred_type = type(inferred_value)

            else:
                inferred_type = type(feature_value)
                inferred_value = feature_value

            fields.append(
                Field(
                    name=feature_name,
                    dtype=from_value_type(
                        python_type_to_feast_value_type(
                            feature_name,
                            value=inferred_value,
                            type_name=inferred_type.__name__ if inferred_type else None,
                        )
                    ),
                )
            )
        return fields

    def __eq__(self, other):
        if not isinstance(other, PythonTransformation):
            raise TypeError(
                "Comparisons should only involve PythonTransformation class objects."
            )

        if (
            self.udf_string != other.udf_string
            or self.udf.__code__.co_code != other.udf.__code__.co_code
        ):
            return False

        return True

    @classmethod
    def from_proto(cls, user_defined_function_proto: UserDefinedFunctionProto):
        return PythonTransformation(
            udf=dill.loads(user_defined_function_proto.body),
            udf_string=user_defined_function_proto.body_text,
        )



import logging
import uuid
from time import sleep
from typing import List, Literal

import pyarrow as pa
import yaml
from kubernetes import client, utils
from kubernetes import config as k8s_config
from kubernetes.client.exceptions import ApiException
from kubernetes.utils import FailToCreateError
from pydantic import StrictStr

from feast import RepoConfig
from feast.infra.common.materialization_job import (
    MaterializationJobStatus,
    MaterializationTask,
)
from feast.infra.common.retrieval_task import HistoricalRetrievalTask
from feast.infra.compute_engines.base import ComputeEngine
from feast.infra.offline_stores.offline_store import OfflineStore
from feast.infra.online_stores.online_store import OnlineStore
from feast.infra.registry.base_registry import BaseRegistry
from feast.repo_config import FeastConfigBaseModel
from feast.utils import _get_column_names

from .k8s_materialization_job import KubernetesMaterializationJob

logger = logging.getLogger(__name__)


class KubernetesComputeEngineConfig(FeastConfigBaseModel):
    """Batch Compute Engine config for Kubernetes"""

    type: Literal["k8s"] = "k8s"
    """ Materialization type selector"""

    namespace: StrictStr = "default"
    """ (optional) The namespace in Kubernetes to use when creating services, configuration maps and jobs.
    """

    image: StrictStr = "feast/feast-k8s-materialization:latest"
    """ (optional) The container image to use when running the materialization job."""

    env: List[dict] = []
    """ (optional) A list of environment variables to set in the created Kubernetes pods.
    These environment variables can be used to reference Kubernetes secrets.
    """

    image_pull_secrets: List[dict] = []
    """ (optional) The secrets to use when pulling the image to run for the materialization job """

    resources: dict = {}
    """ (optional) The resource requests and limits for the materialization containers """

    service_account_name: StrictStr = ""
    """ (optional) The service account name to use when running the job """

    annotations: dict = {}
    """ (optional) Annotations to apply to the job container. Useful for linking the service account to IAM roles, operational metadata, etc  """

    include_security_context_capabilities: bool = True
    """ (optional)  Include security context capabilities in the init and job container spec """

    labels: dict = {}
    """ (optional) additional labels to append to kubernetes objects """

    max_parallelism: int = 10
    """ (optional) Maximum number of pods allowed to run in parallel within a single job"""

    synchronous: bool = False
    """ (optional) If true, wait for materialization for one feature to complete before moving to the next """

    retry_limit: int = 2
    """ (optional) Maximum number of times to retry a materialization worker pod"""

    mini_batch_size: int = 1000
    """ (optional) Number of rows to process per write operation (default 1000)"""

    active_deadline_seconds: int = 86400
    """ (optional) Maximum amount of time a materialization job is allowed to run"""

    job_batch_size: int = 100
    """ (optional) Maximum number of pods to process per job.  Only applies to synchronous materialization"""

    print_pod_logs_on_failure: bool = True
    """(optional) Print pod logs on job failure.  Only applies to synchronous materialization"""


class KubernetesComputeEngine(ComputeEngine):
    def get_historical_features(
        self, registry: BaseRegistry, task: HistoricalRetrievalTask
    ) -> pa.Table:
        raise NotImplementedError(
            "KubernetesComputeEngine does not support get_historical_features()"
        )

    def __init__(
        self,
        *,
        repo_config: RepoConfig,
        offline_store: OfflineStore,
        online_store: OnlineStore,
        **kwargs,
    ):
        super().__init__(
            repo_config=repo_config,
            offline_store=offline_store,
            online_store=online_store,
            **kwargs,
        )
        self.repo_config = repo_config
        self.offline_store = offline_store
        self.online_store = online_store

        k8s_config.load_config()

        self.k8s_client = client.api_client.ApiClient()
        self.v1 = client.CoreV1Api(self.k8s_client)
        self.batch_v1 = client.BatchV1Api(self.k8s_client)
        self.batch_engine_config = repo_config.batch_engine
        self.namespace = self.batch_engine_config.namespace

    def _materialize_one(
        self,
        registry: BaseRegistry,
        task: MaterializationTask,
        **kwargs,
    ):
        feature_view = task.feature_view
        start_date = task.start_time
        end_date = task.end_time
        project = task.project

        entities = []
        for entity_name in feature_view.entities:
            entities.append(registry.get_entity(entity_name, project))

        (
            join_key_columns,
            feature_name_columns,
            timestamp_field,
            created_timestamp_column,
        ) = _get_column_names(feature_view, entities)

        offline_job = self.offline_store.pull_latest_from_table_or_query(
            config=self.repo_config,
            data_source=feature_view.batch_source,
            join_key_columns=join_key_columns,
            feature_name_columns=feature_name_columns,
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            start_date=start_date,
            end_date=end_date,
        )

        paths = offline_job.to_remote_storage()
        if self.batch_engine_config.synchronous:
            offset = 0
            total_pods = len(paths)
            batch_size = self.batch_engine_config.job_batch_size
            if batch_size < 1:
                raise ValueError("job_batch_size must be a value greater than 0")
            if batch_size < self.batch_engine_config.max_parallelism:
                logger.warning(
                    "job_batch_size is less than max_parallelism. Setting job_batch_size = max_parallelism"
                )
                batch_size = self.batch_engine_config.max_parallelism

            while True:
                next_offset = min(offset + batch_size, total_pods)
                job = self._await_path_materialization(
                    paths[offset:next_offset],
                    feature_view,
                    offset,
                    next_offset,
                    total_pods,
                )
                offset += batch_size
                if (
                    offset >= total_pods
                    or job.status() == MaterializationJobStatus.ERROR
                ):
                    break
        else:
            job_id = str(uuid.uuid4())
            job = self._create_kubernetes_job(job_id, paths, feature_view)

        return job

    def _await_path_materialization(
        self, paths, feature_view, batch_start, batch_end, total_pods
    ):
        job_id = str(uuid.uuid4())
        job = self._create_kubernetes_job(job_id, paths, feature_view)

        try:
            while job.status() in (
                MaterializationJobStatus.WAITING,
                MaterializationJobStatus.RUNNING,
            ):
                logger.info(
                    f"{feature_view.name} materialization for pods {batch_start}-{batch_end} "
                    f"(of {total_pods}) running..."
                )
                sleep(30)
            logger.info(
                f"{feature_view.name} materialization for pods {batch_start}-{batch_end} "
                f"(of {total_pods}) complete with status {job.status()}"
            )
        except BaseException as e:
            logger.info(f"Deleting job {job.job_id()}")
            try:
                self.batch_v1.delete_namespaced_job(job.job_id(), self.namespace)
            except ApiException as ae:
                logger.warning(f"Could not delete job due to API Error: {ae.body}")
            raise e
        finally:
            logger.info(f"Deleting configmap {self._configmap_name(job_id)}")
            try:
                self.v1.delete_namespaced_config_map(
                    self._configmap_name(job_id), self.namespace
                )
            except ApiException as ae:
                logger.warning(
                    f"Could not delete configmap due to API Error: {ae.body}"
                )

            if (
                job.status() == MaterializationJobStatus.ERROR
                and self.batch_engine_config.print_pod_logs_on_failure
            ):
                self._print_pod_logs(job.job_id(), feature_view, batch_start)

        return job

    def _print_pod_logs(self, job_id, feature_view, offset=0):
        pods_list = self.v1.list_namespaced_pod(
            namespace=self.namespace,
            label_selector=f"job-name={job_id}",
        ).items
        for i, pod in enumerate(pods_list):
            logger.info(f"Logging output for {feature_view.name} pod {offset + i}")
            try:
                logger.info(
                    self.v1.read_namespaced_pod_log(pod.metadata.name, self.namespace)
                )
            except ApiException as e:
                logger.warning(f"Could not retrieve pod logs due to: {e.body}")

    def _create_kubernetes_job(self, job_id, paths, feature_view):
        try:
            # Create a k8s configmap with information needed by pods
            self._create_configuration_map(job_id, paths, feature_view, self.namespace)

            # Create the k8s job definition
            self._create_job_definition(
                job_id=job_id,
                namespace=self.namespace,
                pods=len(paths),  # Create a pod for each parquet file
                env=self.batch_engine_config.env,
            )
            job = KubernetesMaterializationJob(job_id, self.namespace)
            logger.info(f"Created job `{job.job_id()}` on namespace `{self.namespace}`")
            return job
        except FailToCreateError as failures:
            return KubernetesMaterializationJob(job_id, self.namespace, error=failures)

    def _create_configuration_map(self, job_id, paths, feature_view, namespace):
        """Create a Kubernetes configmap for this job"""

        feature_store_configuration = yaml.dump(
            self.repo_config.model_dump(by_alias=True)
        )

        materialization_config = yaml.dump(
            {"paths": paths, "feature_view": feature_view.name}
        )

        labels = {"feast-materializer": "configmap"}
        configmap_manifest = {
            "kind": "ConfigMap",
            "apiVersion": "v1",
            "metadata": {
                "name": self._configmap_name(job_id),
                "labels": {**labels, **self.batch_engine_config.labels},
            },
            "data": {
                "feature_store.yaml": feature_store_configuration,
                "materialization_config.yaml": materialization_config,
            },
        }
        self.v1.create_namespaced_config_map(
            namespace=namespace,
            body=configmap_manifest,
        )

    def _configmap_name(self, job_id):
        return f"feast-{job_id}"

    def _create_job_definition(self, job_id, namespace, pods, env, index_offset=0):
        """Create a kubernetes job definition."""
        job_env = [
            {
                "name": "MINI_BATCH_SIZE",
                "value": str(self.batch_engine_config.mini_batch_size),
            },
        ]
        # Add any Feast configured environment variables
        job_env.extend(env)

        securityContextCapabilities = None
        if self.batch_engine_config.include_security_context_capabilities:
            securityContextCapabilities = {
                "add": ["NET_BIND_SERVICE"],
                "drop": ["ALL"],
            }

        job_labels = {"feast-materializer": "job"}
        pod_labels = {"feast-materializer": "pod"}
        job_definition = {
            "apiVersion": "batch/v1",
            "kind": "Job",
            "metadata": {
                "name": f"feast-materialization-{job_id}",
                "namespace": namespace,
                "labels": {**job_labels, **self.batch_engine_config.labels},
            },
            "spec": {
                "ttlSecondsAfterFinished": 3600,
                "backoffLimit": self.batch_engine_config.retry_limit,
                "completions": pods,
                "parallelism": min(pods, self.batch_engine_config.max_parallelism),
                "activeDeadlineSeconds": self.batch_engine_config.active_deadline_seconds,
                "completionMode": "Indexed",
                "template": {
                    "metadata": {
                        "annotations": self.batch_engine_config.annotations,
                        "labels": {**pod_labels, **self.batch_engine_config.labels},
                    },
                    "spec": {
                        "restartPolicy": "Never",
                        "subdomain": f"feast-materialization-{job_id}",
                        "imagePullSecrets": self.batch_engine_config.image_pull_secrets,
                        "serviceAccountName": self.batch_engine_config.service_account_name,
                        "containers": [
                            {
                                "command": ["python", "main.py"],
                                "env": job_env,
                                "image": self.batch_engine_config.image,
                                "imagePullPolicy": "Always",
                                "name": "feast",
                                "resources": self.batch_engine_config.resources,
                                "securityContext": {
                                    "allowPrivilegeEscalation": False,
                                    "capabilities": securityContextCapabilities,
                                    "readOnlyRootFilesystem": False,
                                },
                                "terminationMessagePath": "/dev/termination-log",
                                "terminationMessagePolicy": "File",
                                "volumeMounts": [
                                    {
                                        "mountPath": "/var/feast/",
                                        "name": self._configmap_name(job_id),
                                    },
                                ],
                            }
                        ],
                        "volumes": [
                            {
                                "configMap": {
                                    "defaultMode": 420,
                                    "name": self._configmap_name(job_id),
                                },
                                "name": "python-files",
                            },
                            {
                                "configMap": {"name": self._configmap_name(job_id)},
                                "name": self._configmap_name(job_id),
                            },
                        ],
                    },
                },
            },
        }
        utils.create_from_dict(self.k8s_client, job_definition)



"""
Ray Compute Engine for Feast

This module provides a Ray-based compute engine for distributed feature computation.
It includes:
- RayComputeEngine: Main compute engine implementation
- RayComputeEngineConfig: Configuration for the compute engine
- Ray DAG nodes for distributed processing
"""

from .compute import RayComputeEngine
from .config import RayComputeEngineConfig
from .feature_builder import RayFeatureBuilder
from .job import RayDAGRetrievalJob, RayMaterializationJob
from .nodes import (
    RayAggregationNode,
    RayDedupNode,
    RayFilterNode,
    RayJoinNode,
    RayReadNode,
    RayTransformationNode,
    RayWriteNode,
)

__all__ = [
    "RayComputeEngine",
    "RayComputeEngineConfig",
    "RayDAGRetrievalJob",
    "RayMaterializationJob",
    "RayFeatureBuilder",
    "RayReadNode",
    "RayJoinNode",
    "RayFilterNode",
    "RayAggregationNode",
    "RayDedupNode",
    "RayTransformationNode",
    "RayWriteNode",
]



import asyncio
import logging

import grpc

from feast.permissions.auth.auth_manager import (
    get_auth_manager,
)
from feast.permissions.security_manager import get_security_manager

logger = logging.getLogger(__name__)


class AuthInterceptor(grpc.ServerInterceptor):
    def intercept_service(self, continuation, handler_call_details):
        sm = get_security_manager()

        if sm is not None:
            auth_manager = get_auth_manager()
            access_token = auth_manager.token_extractor.extract_access_token(
                metadata=dict(handler_call_details.invocation_metadata)
            )

            logger.debug(
                f"Fetching user details for token of length: {len(access_token)}"
            )
            current_user = asyncio.run(
                auth_manager.token_parser.user_details_from_access_token(access_token)
            )
            logger.debug(f"User is: {current_user}")
            sm.set_current_user(current_user)

        return continuation(handler_call_details)



import logging

from feast.errors import FeastPermissionError
from feast.feast_object import FeastObject
from feast.permissions.decision import DecisionEvaluator
from feast.permissions.permission import (
    AuthzedAction,
    Permission,
)
from feast.permissions.user import User

logger = logging.getLogger(__name__)


def enforce_policy(
    permissions: list[Permission],
    user: User,
    resources: list[FeastObject],
    actions: list[AuthzedAction],
    filter_only: bool = False,
) -> list[FeastObject]:
    """
    Define the logic to apply the configured permissions when a given action is requested on
    a protected resource.

    If no permissions are defined, the result is to deny the execution.

    Args:
        permissions: The configured set of `Permission`.
        user: The current user.
        resources: The resources for which we need to enforce authorized permission.
        actions: The requested actions to be authorized.
        filter_only: If `True`, it removes unauthorized resources from the returned value, otherwise it raises a `FeastPermissionError` the
        first unauthorized resource. Defaults to `False`.

    Returns:
        list[FeastObject]: A filtered list of the permitted resources.

    Raises:
        FeastPermissionError: If the current user is not authorized to eecute the requested actions on the given resources (and `filter_only` is `False`).
    """
    if not permissions:
        # If no permissions are defined, deny access to all resources
        # This is a security measure to prevent unauthorized access
        logger.warning("No permissions defined - denying access to all resources")
        if not filter_only:
            raise FeastPermissionError("No permissions defined - access denied")
        return []

    _permitted_resources: list[FeastObject] = []
    for resource in resources:
        logger.debug(
            f"Enforcing permission policies for {type(resource).__name__}:{resource.name} to execute {actions}"
        )
        matching_permissions = [
            p
            for p in permissions
            if p.match_resource(resource) and p.match_actions(actions)
        ]

        if matching_permissions:
            evaluator = DecisionEvaluator(len(matching_permissions))
            for p in matching_permissions:
                permission_grant, permission_explanation = p.policy.validate_user(
                    user=user
                )
                evaluator.add_grant(
                    permission_grant,
                    f"Permission {p.name} denied execution of {[a.value.upper() for a in actions]} to {type(resource).__name__}:{resource.name}: {permission_explanation}",
                )

                if evaluator.is_decided():
                    grant, explanations = evaluator.grant()
                    if not grant and not filter_only:
                        logger.error(f"Permission denied: {','.join(explanations)}")
                        raise FeastPermissionError(",".join(explanations))
                    if grant:
                        logger.debug(
                            f"Permission granted for {type(resource).__name__}:{resource.name}"
                        )
                        _permitted_resources.append(resource)
                    break
        else:
            message = f"No permissions defined to manage {actions} on {type(resource)}/{resource.name}."
            logger.exception(f"**PERMISSION NOT GRANTED**: {message}")
            raise FeastPermissionError(message)
    return _permitted_resources



# Copyright 2025 The Feast Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file provides integration test repo configuration for HybridOnlineStore.
# It enables running integration tests with multiple online store backends.
# Update this file if you add more backends or change test setup.

from tests.integration.feature_repos.integration_test_repo_config import (
    IntegrationTestRepoConfig,
)
from tests.integration.feature_repos.universal.online_store.hybrid_online_store import (
    HybridOnlineStoreCreator,
)

FULL_REPO_CONFIGS = [
    IntegrationTestRepoConfig(online_store_creator=HybridOnlineStoreCreator),
]



from typing import Any, Dict

import numpy as np
import pandas as pd
import psycopg
import pyarrow as pa
from psycopg import AsyncConnection, Connection
from psycopg.conninfo import make_conninfo
from psycopg_pool import AsyncConnectionPool, ConnectionPool

from feast.infra.utils.postgres.postgres_config import PostgreSQLConfig
from feast.type_map import arrow_to_pg_type


def _get_conn(config: PostgreSQLConfig) -> Connection:
    """Get a psycopg `Connection`."""
    conn = psycopg.connect(
        conninfo=_get_conninfo(config),
        keepalives_idle=config.keepalives_idle,
        **_get_conn_kwargs(config),
    )
    return conn


async def _get_conn_async(config: PostgreSQLConfig) -> AsyncConnection:
    """Get a psycopg `AsyncConnection`."""
    conn = await psycopg.AsyncConnection.connect(
        conninfo=_get_conninfo(config),
        keepalives_idle=config.keepalives_idle,
        **_get_conn_kwargs(config),
    )
    return conn


def _get_connection_pool(config: PostgreSQLConfig) -> ConnectionPool:
    """Get a psycopg `ConnectionPool`."""
    return ConnectionPool(
        conninfo=_get_conninfo(config),
        min_size=config.min_conn,
        max_size=config.max_conn,
        open=False,
        kwargs=_get_conn_kwargs(config),
    )


async def _get_connection_pool_async(config: PostgreSQLConfig) -> AsyncConnectionPool:
    """Get a psycopg `AsyncConnectionPool`."""
    return AsyncConnectionPool(
        conninfo=_get_conninfo(config),
        min_size=config.min_conn,
        max_size=config.max_conn,
        open=False,
        kwargs=_get_conn_kwargs(config),
    )


def _get_conninfo(config: PostgreSQLConfig) -> str:
    """Get the `conninfo` argument required for connection objects."""
    return make_conninfo(
        conninfo="",
        user=config.user,
        password=config.password,
        host=config.host,
        port=int(config.port),
        dbname=config.database,
    )


def _get_conn_kwargs(config: PostgreSQLConfig) -> Dict[str, Any]:
    """Get the additional `kwargs` required for connection objects."""
    return {
        "sslmode": config.sslmode,
        "sslkey": config.sslkey_path,
        "sslcert": config.sslcert_path,
        "sslrootcert": config.sslrootcert_path,
        "options": "-c search_path={}".format(config.db_schema or config.user),
    }


def _df_to_create_table_sql(entity_df, table_name) -> str:
    pa_table = pa.Table.from_pandas(entity_df)
    columns = [
        f""""{f.name}" {arrow_to_pg_type(str(f.type))}""" for f in pa_table.schema
    ]
    return f"""
        CREATE TABLE "{table_name}" (
            {", ".join(columns)}
        );
        """


def df_to_postgres_table(
    config: PostgreSQLConfig, df: pd.DataFrame, table_name: str
) -> Dict[str, np.dtype]:
    """
    Create a table for the data frame, insert all the values, and return the table schema
    """
    nr_columns = df.shape[1]
    placeholders = ", ".join(["%s"] * nr_columns)
    query = f"INSERT INTO {table_name} VALUES ({placeholders})"
    values = df.replace({np.nan: None}).to_numpy().tolist()

    with _get_conn(config) as conn, conn.cursor() as cur:
        cur.execute(_df_to_create_table_sql(df, table_name))
        cur.executemany(query, values)
        return dict(zip(df.columns, df.dtypes))


def get_query_schema(config: PostgreSQLConfig, sql_query: str) -> Dict[str, np.dtype]:
    """
    We'll use the statement when we perform the query rather than copying data to a
    new table
    """
    with _get_conn(config) as conn:
        conn.read_only = True
        df = pd.read_sql(
            f"SELECT * FROM {sql_query} LIMIT 0",
            conn,
        )
        return dict(zip(df.columns, df.dtypes))



"""
A module with utility functions to support authorizing the REST servers using the FastAPI framework.
"""

from typing import Any

from fastapi import HTTPException
from fastapi.requests import Request

from feast.permissions.auth.auth_manager import (
    get_auth_manager,
)
from feast.permissions.security_manager import get_security_manager


async def inject_user_details(request: Request) -> Any:
    """
    A function to extract the authorization token from a user request, extract the user details and propagate them to the
    current security manager, if any.
    """
    sm = get_security_manager()
    current_user = None
    if sm is not None:
        try:
            auth_manager = get_auth_manager()
            access_token = auth_manager.token_extractor.extract_access_token(
                request=request
            )
            if not access_token:
                raise HTTPException(
                    status_code=401, detail="Missing authentication token"
                )

            current_user = (
                await auth_manager.token_parser.user_details_from_access_token(
                    access_token=access_token
                )
            )

            sm.set_current_user(current_user)
        except Exception:
            raise HTTPException(
                status_code=401, detail="Invalid or expired access token"
            )

    return current_user



import json
from typing import Any, Dict, List, Optional

from feast.feature import Feature


class DocumentLabel:
    def __init__(
        self,
        chunk_id: str,
        document_id: str,
        label: str,
        confidence: Optional[float] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.chunk_id = chunk_id
        self.document_id = document_id
        self.label = label
        self.confidence = confidence
        self.metadata = metadata or {}

    def to_dict(self) -> Dict[str, Any]:
        return {
            "chunk_id": self.chunk_id,
            "document_id": self.document_id,
            "label": self.label,
            "confidence": self.confidence,
            "metadata": self.metadata,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DocumentLabel":
        return cls(
            chunk_id=data["chunk_id"],
            document_id=data["document_id"],
            label=data["label"],
            confidence=data.get("confidence"),
            metadata=data.get("metadata", {}),
        )


def store_document_label(feature: Feature, label: DocumentLabel) -> None:
    if not hasattr(feature, "labels") or feature.labels is None:
        if hasattr(feature, "_labels"):
            feature._labels = {}
        else:
            return

    labels_dict = feature.labels if hasattr(feature, "labels") else feature._labels
    labels_key = "document_labels"
    if labels_key not in labels_dict:
        labels_dict[labels_key] = "[]"

    existing_labels = json.loads(labels_dict[labels_key])
    existing_labels.append(label.to_dict())
    labels_dict[labels_key] = json.dumps(existing_labels)


def get_document_labels(feature: Feature) -> List[DocumentLabel]:
    labels_dict = None
    if hasattr(feature, "labels") and feature.labels:
        labels_dict = feature.labels
    elif hasattr(feature, "_labels") and feature._labels:
        labels_dict = feature._labels

    if not labels_dict or "document_labels" not in labels_dict:
        return []

    labels_data = json.loads(labels_dict["document_labels"])
    return [DocumentLabel.from_dict(label_dict) for label_dict in labels_data]


def remove_document_label(feature: Feature, chunk_id: str, document_id: str) -> bool:
    labels_dict = None
    if hasattr(feature, "labels") and feature.labels:
        labels_dict = feature.labels
    elif hasattr(feature, "_labels") and feature._labels:
        labels_dict = feature._labels

    if not labels_dict or "document_labels" not in labels_dict:
        return False

    existing_labels = json.loads(labels_dict["document_labels"])
    original_length = len(existing_labels)

    filtered_labels = [
        label
        for label in existing_labels
        if not (label["chunk_id"] == chunk_id and label["document_id"] == document_id)
    ]

    if len(filtered_labels) < original_length:
        labels_dict["document_labels"] = json.dumps(filtered_labels)
        return True

    return False



import click
import yaml

from feast import utils
from feast.cli.cli_options import tagsOption
from feast.errors import FeastObjectNotFoundException
from feast.feature_view import FeatureView
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.repo_operations import create_feature_store


@click.group(name="feature-views")
def feature_views_cmd():
    """
    Access feature views
    """
    pass


@feature_views_cmd.command("describe")
@click.argument("name", type=click.STRING)
@click.pass_context
def feature_view_describe(ctx: click.Context, name: str):
    """
    Describe a feature view
    """
    store = create_feature_store(ctx)

    try:
        feature_view = store.get_feature_view(name)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(feature_view)), default_flow_style=False, sort_keys=False
        )
    )


@feature_views_cmd.command(name="list")
@tagsOption
@click.pass_context
def feature_view_list(ctx: click.Context, tags: list[str]):
    """
    List all feature views
    """
    store = create_feature_store(ctx)
    table = []
    tags_filter = utils.tags_list_to_dict(tags)
    for feature_view in [
        *store.list_batch_feature_views(tags=tags_filter),
        *store.list_on_demand_feature_views(tags=tags_filter),
    ]:
        entities = set()
        if isinstance(feature_view, FeatureView):
            entities.update(feature_view.entities)
        elif isinstance(feature_view, OnDemandFeatureView):
            for backing_fv in feature_view.source_feature_view_projections.values():
                entities.update(store.get_feature_view(backing_fv.name).entities)
        table.append(
            [
                feature_view.name,
                entities if len(entities) > 0 else "n/a",
                type(feature_view).__name__,
            ]
        )

    from tabulate import tabulate

    print(tabulate(table, headers=["NAME", "ENTITIES", "TYPE"], tablefmt="plain"))



import contextlib
import re
from dataclasses import asdict
from datetime import datetime
from typing import Iterator, List, Literal, Optional, Union, cast

import numpy as np
import pandas as pd
import pyarrow as pa
from pyarrow.compute import cast as pa_cast

from feast import FeatureView, OnDemandFeatureView, RepoConfig
from feast.data_source import DataSource
from feast.errors import EntitySQLEmptyResults, InvalidEntityType
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL
from feast.infra.offline_stores import offline_utils
from feast.infra.offline_stores.contrib.clickhouse_offline_store.clickhouse_source import (
    ClickhouseSource,
    SavedDatasetClickhouseStorage,
)
from feast.infra.offline_stores.contrib.postgres_offline_store.postgres import (
    PostgreSQLRetrievalJob,
    build_point_in_time_query,
)
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.registry.base_registry import BaseRegistry
from feast.infra.utils.clickhouse.clickhouse_config import ClickhouseConfig
from feast.infra.utils.clickhouse.connection_utils import get_client
from feast.saved_dataset import SavedDatasetStorage


class ClickhouseOfflineStoreConfig(ClickhouseConfig):
    type: Literal["clickhouse"] = "clickhouse"


class ClickhouseOfflineStore(OfflineStore):
    @staticmethod
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, ClickhouseOfflineStoreConfig)
        for fv in feature_views:
            assert isinstance(fv.batch_source, ClickhouseSource)

        entity_schema = _get_entity_schema(entity_df, config)

        entity_df_event_timestamp_col = (
            offline_utils.infer_event_timestamp_from_entity_df(entity_schema)
        )

        entity_df_event_timestamp_range = _get_entity_df_event_timestamp_range(
            entity_df,
            entity_df_event_timestamp_col,
            config,
        )

        @contextlib.contextmanager
        def query_generator() -> Iterator[str]:
            table_name = offline_utils.get_temp_entity_table_name()
            if (
                isinstance(entity_df, pd.DataFrame)
                and not config.offline_store.use_temporary_tables_for_entity_df
            ):
                table_name = f"{config.offline_store.database}.{table_name}"

            _upload_entity_df(
                config,
                entity_df,
                table_name,
                entity_df_event_timestamp_col,
            )

            expected_join_keys = offline_utils.get_expected_join_keys(
                project, feature_views, registry
            )

            offline_utils.assert_expected_columns_in_entity_df(
                entity_schema, expected_join_keys, entity_df_event_timestamp_col
            )

            query_context = offline_utils.get_feature_view_query_context(
                feature_refs,
                feature_views,
                registry,
                project,
                entity_df_event_timestamp_range,
            )

            query_context_dict = [asdict(context) for context in query_context]
            # Hack for query_context.entity_selections to support uppercase in columns
            for context in query_context_dict:
                context["entity_selections"] = [
                    f""""{entity_selection.replace(" AS ", '" AS "')}\""""
                    for entity_selection in context["entity_selections"]
                ]

            try:
                query = build_point_in_time_query(
                    query_context_dict,
                    left_table_query_string=table_name,
                    entity_df_event_timestamp_col=entity_df_event_timestamp_col,
                    entity_df_columns=entity_schema.keys(),
                    query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
                    full_feature_names=full_feature_names,
                )
                yield query
            finally:
                if (
                    table_name
                    and not config.offline_store.use_temporary_tables_for_entity_df
                ):
                    get_client(config.offline_store).command(
                        f"DROP TABLE IF EXISTS {table_name}"
                    )

        return ClickhouseRetrievalJob(
            query=query_generator,
            config=config,
            full_feature_names=full_feature_names,
            on_demand_feature_views=OnDemandFeatureView.get_requested_odfvs(
                feature_refs, project, registry
            ),
            metadata=RetrievalMetadata(
                features=feature_refs,
                keys=list(entity_schema.keys() - {entity_df_event_timestamp_col}),
                min_event_timestamp=entity_df_event_timestamp_range[0],
                max_event_timestamp=entity_df_event_timestamp_range[1],
            ),
        )

    @staticmethod
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, ClickhouseOfflineStoreConfig)
        assert isinstance(data_source, ClickhouseSource)
        from_expression = data_source.get_table_query_string()

        partition_by_join_key_string = ", ".join(_append_alias(join_key_columns, "a"))
        if partition_by_join_key_string != "":
            partition_by_join_key_string = (
                "PARTITION BY " + partition_by_join_key_string
            )
        timestamps = [timestamp_field]
        if created_timestamp_column:
            timestamps.append(created_timestamp_column)
        timestamp_desc_string = " DESC, ".join(_append_alias(timestamps, "a")) + " DESC"
        a_field_string = ", ".join(
            _append_alias(join_key_columns + feature_name_columns + timestamps, "a")
        )
        b_field_string = ", ".join(
            _append_alias(join_key_columns + feature_name_columns + timestamps, "b")
        )

        query = f"""
            SELECT
                {b_field_string}
                {f", {repr(DUMMY_ENTITY_VAL)} AS {DUMMY_ENTITY_ID}" if not join_key_columns else ""}
            FROM (
                SELECT {a_field_string},
                ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS _feast_row
                FROM {from_expression} a
                WHERE a."{timestamp_field}"
                    BETWEEN toDateTime64('{start_date.replace(tzinfo=None)!s}', 6, '{start_date.tzinfo!s}')
                    AND toDateTime64('{end_date.replace(tzinfo=None)!s}', 6, '{end_date.tzinfo!s}')
            ) b
            WHERE _feast_row = 1
            """

        return ClickhouseRetrievalJob(
            query=query,
            config=config,
            full_feature_names=False,
            on_demand_feature_views=None,
        )


class ClickhouseRetrievalJob(PostgreSQLRetrievalJob):
    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        with self._query_generator() as query:
            results = get_client(self.config.offline_store).query_df(query)
            return results

    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pa.Table:
        with self._query_generator() as query:
            results: pa.Table = get_client(self.config.offline_store).query_arrow(query)
            # Feast doesn't support native decimal types, so we must convert decimal columns to double
            for col_index, (name, dtype) in enumerate(
                zip(results.schema.names, results.schema.types)
            ):
                if pa.types.is_decimal(dtype):
                    results = results.set_column(
                        col_index,
                        name,
                        pa_cast(results[name], target_type=pa.float64()),
                    )
            return results

    def persist(
        self,
        storage: SavedDatasetStorage,
        allow_overwrite: Optional[bool] = False,
        timeout: Optional[int] = None,
    ):
        assert isinstance(storage, SavedDatasetClickhouseStorage)

        df_to_clickhouse_table(
            config=self.config.offline_store,
            df=self.to_df(),
            table_name=storage.clickhouse_options._table,
            entity_timestamp_col="event_timestamp",
        )


def _get_entity_schema(
    entity_df: Union[pd.DataFrame, str],
    config: RepoConfig,
) -> dict[str, np.dtype]:
    if isinstance(entity_df, pd.DataFrame):
        return dict(zip(entity_df.columns, entity_df.dtypes))
    elif isinstance(entity_df, str):
        query = f"SELECT * FROM ({entity_df}) LIMIT 1"
        df = get_client(config.offline_store).query_df(query)
        return _get_entity_schema(df, config)
    else:
        raise InvalidEntityType(type(entity_df))


def _get_entity_df_event_timestamp_range(
    entity_df: Union[pd.DataFrame, str],
    entity_df_event_timestamp_col: str,
    config: RepoConfig,
) -> tuple[datetime, datetime]:
    if isinstance(entity_df, pd.DataFrame):
        entity_df_event_timestamp = entity_df.loc[
            :, entity_df_event_timestamp_col
        ].infer_objects()
        if pd.api.types.is_string_dtype(entity_df_event_timestamp):
            entity_df_event_timestamp = pd.to_datetime(
                entity_df_event_timestamp, utc=True
            )
        entity_df_event_timestamp_range = (
            entity_df_event_timestamp.min().to_pydatetime(),
            entity_df_event_timestamp.max().to_pydatetime(),
        )
    elif isinstance(entity_df, str):
        # If the entity_df is a string (SQL query), determine range
        # from table
        query = f'SELECT MIN("{entity_df_event_timestamp_col}") AS "min_value", MAX("{entity_df_event_timestamp_col}") AS "max_value" FROM ({entity_df})'
        results = get_client(config.offline_store).query(query).result_rows

        entity_df_event_timestamp_range = cast(tuple[datetime, datetime], results[0])
        if (
            entity_df_event_timestamp_range[0] is None
            or entity_df_event_timestamp_range[1] is None
        ):
            raise EntitySQLEmptyResults(entity_df)
    else:
        raise InvalidEntityType(type(entity_df))

    return entity_df_event_timestamp_range


def _upload_entity_df(
    config: RepoConfig,
    entity_df: Union[pd.DataFrame, str],
    table_name: str,
    entity_timestamp_col: str,
) -> None:
    if isinstance(entity_df, pd.DataFrame):
        df_to_clickhouse_table(
            config.offline_store, entity_df, table_name, entity_timestamp_col
        )
    elif isinstance(entity_df, str):
        if config.offline_store.use_temporary_tables_for_entity_df:
            query = f'CREATE TEMPORARY TABLE "{table_name}" AS ({entity_df})'
        else:
            query = f'CREATE TABLE "{table_name}" ENGINE = MergeTree() ORDER BY ({entity_timestamp_col}) AS ({entity_df})'
        get_client(config.offline_store).command(query)
    else:
        raise InvalidEntityType(type(entity_df))


def df_to_clickhouse_table(
    config: ClickhouseConfig,
    df: pd.DataFrame,
    table_name: str,
    entity_timestamp_col: str,
) -> None:
    table_schema = _df_to_create_table_schema(df)
    if config.use_temporary_tables_for_entity_df:
        query = f"CREATE TEMPORARY TABLE {table_name} ({table_schema})"
    else:
        query = f"""
            CREATE TABLE {table_name} (
                {table_schema}
            )
            ENGINE = MergeTree()
            ORDER BY ({entity_timestamp_col})
        """
    get_client(config).command(query)
    get_client(config).insert_df(table_name, df)


def _df_to_create_table_schema(entity_df: pd.DataFrame) -> str:
    pa_table = pa.Table.from_pandas(entity_df)
    columns = [
        f""""{f.name}" {arrow_to_ch_type(str(f.type), f.nullable)}"""
        for f in pa_table.schema
    ]
    return ", ".join(columns)


def arrow_to_ch_type(t_str: str, nullable: bool) -> str:
    list_pattern = r"list<item: (.*)>"
    list_res = re.search(list_pattern, t_str)
    if list_res is not None:
        item_type_str = list_res.group(1)
        return f"Array({arrow_to_ch_type(item_type_str, nullable)})"

    if nullable:
        return f"Nullable({arrow_to_ch_type(t_str, nullable=False)})"

    try:
        if t_str.startswith("timestamp"):
            return _arrow_to_ch_timestamp_type(t_str)
        return {
            "bool": "Boolean",
            "int8": "Int8",
            "int16": "Int16",
            "int32": "Int32",
            "int64": "Int64",
            "uint8": "UInt8",
            "uint16": "UInt16",
            "uint32": "UInt32",
            "uint64": "Uint64",
            "float": "Float32",
            "double": "Float64",
            "string": "String",
        }[t_str]
    except KeyError:
        raise ValueError(f"Unsupported type: {t_str}")


def _arrow_to_ch_timestamp_type(t_str: str) -> str:
    _ARROW_PRECISION_TO_CH_PRECISION = {
        "s": 0,
        "ms": 3,
        "us": 6,
        "ns": 9,
    }

    unit, *rest = t_str.removeprefix("timestamp[").removesuffix("]").split(",")

    unit = unit.strip()
    precision = _ARROW_PRECISION_TO_CH_PRECISION[unit]

    if len(rest):
        tz = rest[0]
        tz = (
            tz.strip()
            .removeprefix("tz=")
            .translate(
                str.maketrans(
                    {  # type: ignore[arg-type]
                        "'": None,
                        '"': None,
                    }
                )
            )
        )
    else:
        tz = None

    if precision > 0:
        if tz is not None:
            return f"DateTime64({precision}, '{tz}')"
        else:
            return f"DateTime64({precision})"
    else:
        if tz is not None:
            return f"DateTime('{tz}')"
        else:
            return "DateTime"


def _append_alias(field_names: List[str], alias: str) -> List[str]:
    return [f'{alias}."{field_name}"' for field_name in field_names]


MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN = """
/*
 Compute a deterministic hash for the `left_table_query_string` that will be used throughout
 all the logic as the field to GROUP BY the data
*/
WITH entity_dataframe AS (
    SELECT *,
        {{entity_df_event_timestamp_col}} AS entity_timestamp
        {% for featureview in featureviews %}
            {% if featureview.entities %}
            ,(
                {% for entity in featureview.entities %}
                    CAST("{{entity}}" as VARCHAR) ||
                {% endfor %}
                CAST("{{entity_df_event_timestamp_col}}" AS VARCHAR)
            ) AS "{{featureview.name}}__entity_row_unique_id"
            {% else %}
            ,CAST("{{entity_df_event_timestamp_col}}" AS VARCHAR) AS "{{featureview.name}}__entity_row_unique_id"
            {% endif %}
        {% endfor %}
    FROM {{ left_table_query_string }}
),

{% for featureview in featureviews %}

"{{ featureview.name }}__entity_dataframe" AS (
    SELECT
        {% if featureview.entities %}"{{ featureview.entities | join('", "') }}",{% endif %}
        entity_timestamp,
        "{{featureview.name}}__entity_row_unique_id"
    FROM entity_dataframe
    GROUP BY
        {% if featureview.entities %}"{{ featureview.entities | join('", "')}}",{% endif %}
        entity_timestamp,
        "{{featureview.name}}__entity_row_unique_id"
),

/*
 This query template performs the point-in-time correctness join for a single feature set table
 to the provided entity table.

 1. We first join the current feature_view to the entity dataframe that has been passed.
 This JOIN has the following logic:
    - For each row of the entity dataframe, only keep the rows where the `timestamp_field`
    is less than the one provided in the entity dataframe
    - If there a TTL for the current feature_view, also keep the rows where the `timestamp_field`
    is higher the the one provided minus the TTL
    - For each row, Join on the entity key and retrieve the `entity_row_unique_id` that has been
    computed previously

 The output of this CTE will contain all the necessary information and already filtered out most
 of the data that is not relevant.
*/

"{{ featureview.name }}__subquery" AS (
    SELECT
        "{{ featureview.timestamp_field }}" as event_timestamp,
        {{ '"' ~ featureview.created_timestamp_column ~ '" as created_timestamp,' if featureview.created_timestamp_column else '' }}
        {{ featureview.entity_selections | join(', ')}}{% if featureview.entity_selections %},{% else %}{% endif %}
        {% for feature in featureview.features %}
            "{{ feature }}" as {% if full_feature_names %}"{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}"{% else %}"{{ featureview.field_mapping.get(feature, feature) }}"{% endif %}{% if loop.last %}{% else %}, {% endif %}
        {% endfor %}
    FROM {{ featureview.table_subquery }} AS sub
    WHERE "{{ featureview.timestamp_field }}" <= (SELECT MAX(entity_timestamp) FROM entity_dataframe)
    {% if featureview.ttl == 0 %}{% else %}
    AND "{{ featureview.timestamp_field }}" >= (SELECT MIN(entity_timestamp) FROM entity_dataframe) - interval {{ featureview.ttl }} second
    {% endif %}
),

"{{ featureview.name }}__base" AS (
    SELECT
        subquery.*,
        entity_dataframe.entity_timestamp,
        entity_dataframe."{{featureview.name}}__entity_row_unique_id"
    FROM "{{ featureview.name }}__subquery" AS subquery
    INNER JOIN "{{ featureview.name }}__entity_dataframe" AS entity_dataframe
    ON TRUE
        {% for entity in featureview.entities %}
        AND subquery."{{ entity }}" = entity_dataframe."{{ entity }}"
        {% endfor %}
    WHERE TRUE
        AND subquery.event_timestamp <= entity_dataframe.entity_timestamp

        {% if featureview.ttl == 0 %}{% else %}
        AND subquery.event_timestamp >= entity_dataframe.entity_timestamp - interval {{ featureview.ttl }} second
        {% endif %}
),

/*
 2. If the `created_timestamp_column` has been set, we need to
 deduplicate the data first. This is done by calculating the
 `MAX(created_at_timestamp)` for each event_timestamp.
 We then join the data on the next CTE
*/
{% if featureview.created_timestamp_column %}
"{{ featureview.name }}__dedup" AS (
    SELECT
        "{{featureview.name}}__entity_row_unique_id",
        event_timestamp,
        MAX(created_timestamp) as created_timestamp
    FROM "{{ featureview.name }}__base"
    GROUP BY "{{featureview.name}}__entity_row_unique_id", event_timestamp
),
{% endif %}

/*
 3. The data has been filtered during the first CTE "*__base"
 Thus we only need to compute the latest timestamp of each feature.
*/
"{{ featureview.name }}__latest" AS (
    SELECT
        event_timestamp,
        {% if featureview.created_timestamp_column %}created_timestamp,{% endif %}
        "{{featureview.name}}__entity_row_unique_id"
    FROM
    (
        SELECT *,
            ROW_NUMBER() OVER(
                PARTITION BY "{{featureview.name}}__entity_row_unique_id"
                ORDER BY event_timestamp DESC{% if featureview.created_timestamp_column %},created_timestamp DESC{% endif %}
            ) AS row_number
        FROM "{{ featureview.name }}__base"
        {% if featureview.created_timestamp_column %}
            INNER JOIN "{{ featureview.name }}__dedup"
            USING ("{{featureview.name}}__entity_row_unique_id", event_timestamp, created_timestamp)
        {% endif %}
    ) AS sub
    WHERE row_number = 1
),

/*
 4. Once we know the latest value of each feature for a given timestamp,
 we can join again the data back to the original "base" dataset
*/
"{{ featureview.name }}__cleaned" AS (
    SELECT base.*
    FROM "{{ featureview.name }}__base" as base
    INNER JOIN "{{ featureview.name }}__latest"
    USING(
        "{{featureview.name}}__entity_row_unique_id",
        event_timestamp
        {% if featureview.created_timestamp_column %}
            ,created_timestamp
        {% endif %}
    )
){% if loop.last %}{% else %}, {% endif %}


{% endfor %}
/*
 Joins the outputs of multiple time travel joins to a single table.
 The entity_dataframe dataset being our source of truth here.
 */

SELECT "{{ final_output_feature_names | join('", "')}}"
FROM entity_dataframe
{% for featureview in featureviews %}
LEFT JOIN (
    SELECT
        "{{featureview.name}}__entity_row_unique_id"
        {% for feature in featureview.features %}
            ,"{% if full_feature_names %}{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}"
        {% endfor %}
    FROM "{{ featureview.name }}__cleaned"
) AS "{{featureview.name}}" USING ("{{featureview.name}}__entity_row_unique_id")
{% endfor %}
"""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/Aggregation.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1c\x66\x65\x61st/core/Aggregation.proto\x12\nfeast.core\x1a\x1egoogle/protobuf/duration.proto\"\x92\x01\n\x0b\x41ggregation\x12\x0e\n\x06\x63olumn\x18\x01 \x01(\t\x12\x10\n\x08\x66unction\x18\x02 \x01(\t\x12.\n\x0btime_window\x18\x03 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x31\n\x0eslide_interval\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationBU\n\x10\x66\x65\x61st.proto.coreB\x10\x41ggregationProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.Aggregation_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\020AggregationProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_AGGREGATION']._serialized_start=77
  _globals['_AGGREGATION']._serialized_end=223
# @@protoc_insertion_point(module_scope)



import requests
from requests import Session

from feast.permissions.auth.auth_type import AuthType
from feast.permissions.auth_model import (
    AuthConfig,
)
from feast.permissions.client.client_auth_token import get_auth_token


class AuthenticatedRequestsSession(Session):
    def __init__(self, auth_token: str):
        super().__init__()
        self.headers.update({"Authorization": f"Bearer {auth_token}"})


def get_http_auth_requests_session(auth_config: AuthConfig) -> Session:
    if auth_config.type == AuthType.NONE.value:
        request_session = requests.session()
    else:
        request_session = AuthenticatedRequestsSession(get_auth_token(auth_config))
    return request_session



import logging
from typing import Dict

from fastapi import APIRouter, Depends, Query

from feast.api.registry.rest.codegen_utils import (
    render_data_source_code,
    render_push_source_code,
    render_request_source_code,
)
from feast.api.registry.rest.rest_utils import (
    aggregate_across_projects,
    create_grpc_pagination_params,
    create_grpc_sorting_params,
    get_object_relationships,
    get_pagination_params,
    get_relationships_for_objects,
    get_sorting_params,
    grpc_call,
    parse_tags,
)
from feast.protos.feast.registry import RegistryServer_pb2
from feast.type_map import _convert_value_type_str_to_value_type
from feast.types import from_value_type

logger = logging.getLogger(__name__)


def get_data_source_router(grpc_handler) -> APIRouter:
    router = APIRouter()

    @router.get("/data_sources")
    def list_data_sources(
        project: str = Query(...),
        include_relationships: bool = Query(
            False, description="Include relationships for each data source"
        ),
        allow_cache: bool = Query(default=True),
        tags: Dict[str, str] = Depends(parse_tags),
        pagination_params: dict = Depends(get_pagination_params),
        sorting_params: dict = Depends(get_sorting_params),
    ):
        req = RegistryServer_pb2.ListDataSourcesRequest(
            project=project,
            allow_cache=allow_cache,
            tags=tags,
            pagination=create_grpc_pagination_params(pagination_params),
            sorting=create_grpc_sorting_params(sorting_params),
        )
        response = grpc_call(grpc_handler.ListDataSources, req)
        data_sources = response.get("dataSources", [])

        result = {
            "dataSources": data_sources,
            "pagination": response.get("pagination", {}),
        }

        if include_relationships:
            relationships = get_relationships_for_objects(
                grpc_handler, data_sources, "dataSource", project, allow_cache
            )
            result["relationships"] = relationships

        return result

    @router.get("/data_sources/all")
    def list_data_sources_all(
        allow_cache: bool = Query(default=True),
        page: int = Query(1, ge=1),
        limit: int = Query(50, ge=1, le=100),
        sort_by: str = Query(None),
        sort_order: str = Query("asc"),
        include_relationships: bool = Query(
            False, description="Include relationships for each data source"
        ),
    ):
        return aggregate_across_projects(
            grpc_handler=grpc_handler,
            list_method=grpc_handler.ListDataSources,
            request_cls=RegistryServer_pb2.ListDataSourcesRequest,
            response_key="dataSources",
            object_type="dataSource",
            allow_cache=allow_cache,
            page=page,
            limit=limit,
            sort_by=sort_by,
            sort_order=sort_order,
            include_relationships=include_relationships,
        )

    @router.get("/data_sources/{name}")
    def get_data_source(
        name: str,
        project: str = Query(...),
        include_relationships: bool = Query(
            False, description="Include relationships for this data source"
        ),
        allow_cache: bool = Query(default=True),
    ):
        req = RegistryServer_pb2.GetDataSourceRequest(
            name=name,
            project=project,
            allow_cache=allow_cache,
        )
        data_source = grpc_call(grpc_handler.GetDataSource, req)

        result = data_source

        if include_relationships:
            relationships = get_object_relationships(
                grpc_handler, "dataSource", name, project, allow_cache
            )
            result["relationships"] = relationships

        if result:
            spec = result.get("spec", result)
            name = spec.get("name", result.get("name", "default_source"))
            source_type = spec.get("type", "").upper()
            if source_type == "REQUEST_SOURCE":
                schema_fields = []
                feast_types = set()
                for field in spec.get("requestDataOptions", {}).get("schema", []):
                    value_type_enum = _convert_value_type_str_to_value_type(
                        field.get("valueType", "").upper()
                    )
                    feast_type = from_value_type(value_type_enum)
                    dtype = getattr(feast_type, "__name__", str(feast_type))
                    schema_fields.append(
                        f'        Field(name="{field["name"]}", dtype={dtype}),'
                    )
                    feast_types.add(dtype)
                context = dict(
                    name=name,
                    schema_lines=schema_fields,
                    feast_types=list(feast_types),
                )
                result["featureDefinition"] = render_request_source_code(context)
                return result
            elif source_type == "PUSH_SOURCE" and "batchSource" in spec:
                batch_source_name = spec["batchSource"].get("name", "batch_source")
                context = dict(
                    name=name,
                    batch_source_name=batch_source_name,
                )
                result["featureDefinition"] = render_push_source_code(context)
                return result
            else:
                path = spec.get("path") or spec.get("fileOptions", {}).get("uri", "")
                timestamp_field = spec.get("timestampField", "event_timestamp")
                created_timestamp_column = spec.get("createdTimestampColumn", "")
                context = dict(
                    name=name,
                    path=path,
                    timestamp_field=timestamp_field,
                    created_timestamp_column=created_timestamp_column,
                )
                result["featureDefinition"] = render_data_source_code(context)
                return result

    return router



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/storage/Redis.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from feast.protos.feast.types import Value_pb2 as feast_dot_types_dot_Value__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x66\x65\x61st/storage/Redis.proto\x12\rfeast.storage\x1a\x17\x66\x65\x61st/types/Value.proto\"^\n\nRedisKeyV2\x12\x0f\n\x07project\x18\x01 \x01(\t\x12\x14\n\x0c\x65ntity_names\x18\x02 \x03(\t\x12)\n\rentity_values\x18\x03 \x03(\x0b\x32\x12.feast.types.ValueBU\n\x13\x66\x65\x61st.proto.storageB\nRedisProtoZ2github.com/feast-dev/feast/go/protos/feast/storageb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.storage.Redis_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\023feast.proto.storageB\nRedisProtoZ2github.com/feast-dev/feast/go/protos/feast/storage'
  _globals['_REDISKEYV2']._serialized_start=69
  _globals['_REDISKEYV2']._serialized_end=163
# @@protoc_insertion_point(module_scope)
